/*
 * Copyright (C) 2016, 2020 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if OS(LINUX)

#include <wtf/NeverDestroyed.h>
#include <wtf/Noncopyable.h>
#include <wtf/text/CString.h>

namespace IPC {
class Attachment;
}

namespace WebKit {

class MemoryPressureMonitor {
    WTF_MAKE_NONCOPYABLE(MemoryPressureMonitor);
    friend NeverDestroyed<MemoryPressureMonitor>;
public:
    static MemoryPressureMonitor& singleton();
    void start();
    static bool disabled();

    ~MemoryPressureMonitor();

private:
    MemoryPressureMonitor() = default;
    bool m_started { false };
    static bool s_disabled;
};

class CGroupMemoryController {
public:
    CGroupMemoryController() = default;
    bool isActive() { return !m_cgroupMemoryControllerPath.isNull(); };

    void setMemoryControllerPath(CString);

    size_t getMemoryTotalWithCgroup();
    size_t getMemoryUsageWithCgroup();

    ~CGroupMemoryController()
    {
        disposeMemoryController();
    }

private:
    CString m_cgroupMemoryControllerPath;

    FILE* m_cgroupMemoryMemswLimitInBytesFile;
    FILE* m_cgroupMemoryLimitInBytesFile;
    FILE* m_cgroupMemoryUsageInBytesFile;

    FILE* m_cgroupV2MemoryMemswMaxFile;
    FILE* m_cgroupV2MemoryMaxFile;
    FILE* m_cgroupV2MemoryHighFile;
    FILE* m_cgroupV2MemoryCurrentFile;

    void disposeMemoryController();
    size_t getCgroupFileValue(FILE*);
};

} // namespace WebKit

#endif // OS(LINUX)
