/*
 * Copyright (C) 2019 Apple Inc. All rights reserved.
 * Copyright (C) 2019, 2021 Igalia S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ScrollingTreeScrollingNodeDelegate.h"

#if ENABLE(ASYNC_SCROLLING) && USE(NICOSIA)

#include "NicosiaPlatformLayer.h"
#include "ScrollingEffectsController.h"
#include "ScrollingStateOverflowScrollingNode.h"
#include "ThreadedScrollingTree.h"
#include <wtf/RunLoop.h>

namespace WebCore {

class ScrollingTreeScrollingNodeDelegateNicosia : public ScrollingTreeScrollingNodeDelegate, public ScrollingEffectsControllerClient {
public:
    explicit ScrollingTreeScrollingNodeDelegateNicosia(ScrollingTreeScrollingNode&, bool scrollAnimatorEnabled);
    virtual ~ScrollingTreeScrollingNodeDelegateNicosia();

    void updateFromStateNode(const ScrollingStateScrollingNode&);
    void updateVisibleLengths();
    WheelEventHandlingResult handleWheelEvent(const PlatformWheelEvent&, EventTargeting);

    // ScrollingTreeScrollingNodeDelegate
    bool startAnimatedScrollToPosition(FloatPoint) final;
    void stopAnimatedScroll() final;

    void serviceScrollAnimation(MonotonicTime) final;

private:
    // ScrollingEffectsControllerClient.
    std::unique_ptr<ScrollingEffectsControllerTimer> createTimer(Function<void()>&&) final;

    void startAnimationCallback(ScrollingEffectsController&) final;
    void stopAnimationCallback(ScrollingEffectsController&) final;

    bool allowsHorizontalScrolling() const final;
    bool allowsVerticalScrolling() const final;

    void immediateScrollBy(const FloatSize&, ScrollClamping = ScrollClamping::Clamped) final;

    void adjustScrollPositionToBoundsIfNecessary() final;

    FloatPoint scrollOffset() const final;

    void willStartScrollSnapAnimation() final;
    void didStopScrollSnapAnimation() final;

    void didStopAnimatedScroll() final;

    float pageScaleFactor() const final;
    ScrollExtents scrollExtents() const final;

    bool scrollAnimationEnabled() const final { return m_scrollAnimatorEnabled; }

    ScrollingEffectsController m_scrollController;

    bool m_scrollAnimatorEnabled { false };
};

} // namespace WebCore

#endif // ENABLE(ASYNC_SCROLLING) && USE(NICOSIA)
