/*
 * Copyright (C) 2021-2022 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(GPU_PROCESS)

#include "WebGPUColor.h"
#include "WebGPUIdentifier.h"
#include <optional>
#include <pal/graphics/WebGPU/WebGPULoadOp.h>
#include <pal/graphics/WebGPU/WebGPUStoreOp.h>
#include <variant>
#include <wtf/Ref.h>
#include <wtf/Vector.h>

namespace WebKit::WebGPU {

struct RenderPassColorAttachment {
    WebGPUIdentifier view;
    WebGPUIdentifier resolveTarget;

    std::optional<Color> clearValue;
    PAL::WebGPU::LoadOp loadOp { PAL::WebGPU::LoadOp::Load };
    PAL::WebGPU::StoreOp storeOp { PAL::WebGPU::StoreOp::Store };

    template<class Encoder> void encode(Encoder& encoder) const
    {
        encoder << view;
        encoder << resolveTarget;
        encoder << clearValue;
        encoder << loadOp;
        encoder << storeOp;
    }

    template<class Decoder> static std::optional<RenderPassColorAttachment> decode(Decoder& decoder)
    {
        std::optional<WebGPUIdentifier> view;
        decoder >> view;
        if (!view)
            return std::nullopt;

        std::optional<WebGPUIdentifier> resolveTarget;
        decoder >> resolveTarget;
        if (!resolveTarget)
            return std::nullopt;

        std::optional<std::optional<Color>> clearValue;
        decoder >> clearValue;
        if (!clearValue)
            return std::nullopt;

        std::optional<PAL::WebGPU::LoadOp> loadOp;
        decoder >> loadOp;
        if (!loadOp)
            return std::nullopt;

        std::optional<PAL::WebGPU::StoreOp> storeOp;
        decoder >> storeOp;
        if (!storeOp)
            return std::nullopt;

        return { { WTFMove(*view), WTFMove(*resolveTarget), WTFMove(*clearValue), WTFMove(*loadOp), WTFMove(*storeOp) } };
    }
};

} // namespace WebKit::WebGPU

#endif // ENABLE(GPU_PROCESS)
