/*
    Copyright (C) 2008 Nokia Corporation and/or its subsidiary(-ies)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

[
    GenerateIsReachable=ReachableFromDOMWindow,
    JSCustomMarkFunction,
    Exposed=Window
] interface Navigator {
    // Non-standard.
    [Conditional=NAVIGATOR_STANDALONE] readonly attribute boolean standalone;

    readonly attribute unsigned long long hardwareConcurrency;

    // Non-standard.
    // FIXME: Remove this. WebKit is the only engine that still has this.
    undefined getStorageUpdates();
};

Navigator includes NavigatorID;
Navigator includes NavigatorLanguage;
Navigator includes NavigatorLocks;
Navigator includes NavigatorMaxTouchPoints;
Navigator includes NavigatorOnLine;
Navigator includes NavigatorCookies;
Navigator includes NavigatorPlugins;
Navigator includes NavigatorServiceWorker;
Navigator includes NavigatorShare;
Navigator includes NavigatorStorage;
Navigator includes NavigatorGPU;
