/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "GPUComputePassTimestampLocation.h"
#include "GPUIntegralTypes.h"
#include "GPUQuerySet.h"
#include <pal/graphics/WebGPU/WebGPUComputePassTimestampWrites.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>

namespace WebCore {

struct GPUComputePassTimestampWrite {
    PAL::WebGPU::ComputePassTimestampWrite convertToBacking() const
    {
        ASSERT(querySet);
        return {
            querySet->backing(),
            queryIndex,
            WebCore::convertToBacking(location),
        };
    }

    GPUQuerySet* querySet { nullptr };
    GPUSize32 queryIndex { 0 };
    GPUComputePassTimestampLocation location { GPUComputePassTimestampLocation::Beginning };
};

using GPUComputePassTimestampWrites = Vector<GPUComputePassTimestampWrite>;

inline PAL::WebGPU::ComputePassTimestampWrites convertToBacking(const GPUComputePassTimestampWrites& computePassTimestampWrites)
{
    return computePassTimestampWrites.map([] (auto& computePassTimestampWrite) {
        return computePassTimestampWrite.convertToBacking();
    });
}

}
