/*
 * Copyright (C) 2018-2020 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

body[dir] .resource-security > section:matches(.connection, .certificate) > .details {
    border-color: var(--network-dns-color);
}

.resource-security .details .key {
    color: var(--network-dns-color);
}

.resource-security .details > .show-certificate > .key {
    display: inline-flex;
    align-items: center;
}

.resource-security .details > .show-certificate > .key > .error {
    width: 14px;
    height: 14px;
    margin-inline-start: 2px;
}

.resource-security .dns-name + .dns-name > .key,
.resource-security .ip-address + .ip-address > .key {
    color: transparent;
}

.resource-security .show-more {
    text-decoration: underline;
    cursor: pointer;
    color: var(--text-color);
}

@media (prefers-color-scheme: dark) {
    body[dir] .resource-security > section:matches(.connection, .certificate) > .details {
        border-color: var(--network-pseudo-header-color);
    }

    .resource-security .details .key {
        color: var(--network-pseudo-header-color);
    }
}
