static const ZydisShortString STR_MNEMONIC[] =
{
    ZYDIS_MAKE_SHORTSTRING("invalid"),
    ZYDIS_MAKE_SHORTSTRING("aaa"),
    ZYDIS_MAKE_SHORTSTRING("aad"),
    ZYDIS_MAKE_SHORTSTRING("aam"),
    ZYDIS_MAKE_SHORTSTRING("aas"),
    ZYDIS_MAKE_SHORTSTRING("adc"),
    ZYDIS_MAKE_SHORTSTRING("adcx"),
    ZYDIS_MAKE_SHORTSTRING("add"),
    ZYDIS_MAKE_SHORTSTRING("addpd"),
    ZYDIS_MAKE_SHORTSTRING("addps"),
    ZYDIS_MAKE_SHORTSTRING("addsd"),
    ZYDIS_MAKE_SHORTSTRING("addss"),
    ZYDIS_MAKE_SHORTSTRING("addsubpd"),
    ZYDIS_MAKE_SHORTSTRING("addsubps"),
    ZYDIS_MAKE_SHORTSTRING("adox"),
    ZYDIS_MAKE_SHORTSTRING("aesdec"),
    ZYDIS_MAKE_SHORTSTRING("aesdec128kl"),
    ZYDIS_MAKE_SHORTSTRING("aesdec256kl"),
    ZYDIS_MAKE_SHORTSTRING("aesdeclast"),
    ZYDIS_MAKE_SHORTSTRING("aesdecwide128kl"),
    ZYDIS_MAKE_SHORTSTRING("aesdecwide256kl"),
    ZYDIS_MAKE_SHORTSTRING("aesenc"),
    ZYDIS_MAKE_SHORTSTRING("aesenc128kl"),
    ZYDIS_MAKE_SHORTSTRING("aesenc256kl"),
    ZYDIS_MAKE_SHORTSTRING("aesenclast"),
    ZYDIS_MAKE_SHORTSTRING("aesencwide128kl"),
    ZYDIS_MAKE_SHORTSTRING("aesencwide256kl"),
    ZYDIS_MAKE_SHORTSTRING("aesimc"),
    ZYDIS_MAKE_SHORTSTRING("aeskeygenassist"),
    ZYDIS_MAKE_SHORTSTRING("and"),
    ZYDIS_MAKE_SHORTSTRING("andn"),
    ZYDIS_MAKE_SHORTSTRING("andnpd"),
    ZYDIS_MAKE_SHORTSTRING("andnps"),
    ZYDIS_MAKE_SHORTSTRING("andpd"),
    ZYDIS_MAKE_SHORTSTRING("andps"),
    ZYDIS_MAKE_SHORTSTRING("arpl"),
    ZYDIS_MAKE_SHORTSTRING("bextr"),
    ZYDIS_MAKE_SHORTSTRING("blcfill"),
    ZYDIS_MAKE_SHORTSTRING("blci"),
    ZYDIS_MAKE_SHORTSTRING("blcic"),
    ZYDIS_MAKE_SHORTSTRING("blcmsk"),
    ZYDIS_MAKE_SHORTSTRING("blcs"),
    ZYDIS_MAKE_SHORTSTRING("blendpd"),
    ZYDIS_MAKE_SHORTSTRING("blendps"),
    ZYDIS_MAKE_SHORTSTRING("blendvpd"),
    ZYDIS_MAKE_SHORTSTRING("blendvps"),
    ZYDIS_MAKE_SHORTSTRING("blsfill"),
    ZYDIS_MAKE_SHORTSTRING("blsi"),
    ZYDIS_MAKE_SHORTSTRING("blsic"),
    ZYDIS_MAKE_SHORTSTRING("blsmsk"),
    ZYDIS_MAKE_SHORTSTRING("blsr"),
    ZYDIS_MAKE_SHORTSTRING("bndcl"),
    ZYDIS_MAKE_SHORTSTRING("bndcn"),
    ZYDIS_MAKE_SHORTSTRING("bndcu"),
    ZYDIS_MAKE_SHORTSTRING("bndldx"),
    ZYDIS_MAKE_SHORTSTRING("bndmk"),
    ZYDIS_MAKE_SHORTSTRING("bndmov"),
    ZYDIS_MAKE_SHORTSTRING("bndstx"),
    ZYDIS_MAKE_SHORTSTRING("bound"),
    ZYDIS_MAKE_SHORTSTRING("bsf"),
    ZYDIS_MAKE_SHORTSTRING("bsr"),
    ZYDIS_MAKE_SHORTSTRING("bswap"),
    ZYDIS_MAKE_SHORTSTRING("bt"),
    ZYDIS_MAKE_SHORTSTRING("btc"),
    ZYDIS_MAKE_SHORTSTRING("btr"),
    ZYDIS_MAKE_SHORTSTRING("bts"),
    ZYDIS_MAKE_SHORTSTRING("bzhi"),
    ZYDIS_MAKE_SHORTSTRING("call"),
    ZYDIS_MAKE_SHORTSTRING("cbw"),
    ZYDIS_MAKE_SHORTSTRING("cdq"),
    ZYDIS_MAKE_SHORTSTRING("cdqe"),
    ZYDIS_MAKE_SHORTSTRING("clac"),
    ZYDIS_MAKE_SHORTSTRING("clc"),
    ZYDIS_MAKE_SHORTSTRING("cld"),
    ZYDIS_MAKE_SHORTSTRING("cldemote"),
    ZYDIS_MAKE_SHORTSTRING("clevict0"),
    ZYDIS_MAKE_SHORTSTRING("clevict1"),
    ZYDIS_MAKE_SHORTSTRING("clflush"),
    ZYDIS_MAKE_SHORTSTRING("clflushopt"),
    ZYDIS_MAKE_SHORTSTRING("clgi"),
    ZYDIS_MAKE_SHORTSTRING("cli"),
    ZYDIS_MAKE_SHORTSTRING("clrssbsy"),
    ZYDIS_MAKE_SHORTSTRING("clts"),
    ZYDIS_MAKE_SHORTSTRING("clui"),
    ZYDIS_MAKE_SHORTSTRING("clwb"),
    ZYDIS_MAKE_SHORTSTRING("clzero"),
    ZYDIS_MAKE_SHORTSTRING("cmc"),
    ZYDIS_MAKE_SHORTSTRING("cmovb"),
    ZYDIS_MAKE_SHORTSTRING("cmovbe"),
    ZYDIS_MAKE_SHORTSTRING("cmovl"),
    ZYDIS_MAKE_SHORTSTRING("cmovle"),
    ZYDIS_MAKE_SHORTSTRING("cmovnb"),
    ZYDIS_MAKE_SHORTSTRING("cmovnbe"),
    ZYDIS_MAKE_SHORTSTRING("cmovnl"),
    ZYDIS_MAKE_SHORTSTRING("cmovnle"),
    ZYDIS_MAKE_SHORTSTRING("cmovno"),
    ZYDIS_MAKE_SHORTSTRING("cmovnp"),
    ZYDIS_MAKE_SHORTSTRING("cmovns"),
    ZYDIS_MAKE_SHORTSTRING("cmovnz"),
    ZYDIS_MAKE_SHORTSTRING("cmovo"),
    ZYDIS_MAKE_SHORTSTRING("cmovp"),
    ZYDIS_MAKE_SHORTSTRING("cmovs"),
    ZYDIS_MAKE_SHORTSTRING("cmovz"),
    ZYDIS_MAKE_SHORTSTRING("cmp"),
    ZYDIS_MAKE_SHORTSTRING("cmppd"),
    ZYDIS_MAKE_SHORTSTRING("cmpps"),
    ZYDIS_MAKE_SHORTSTRING("cmpsb"),
    ZYDIS_MAKE_SHORTSTRING("cmpsd"),
    ZYDIS_MAKE_SHORTSTRING("cmpsq"),
    ZYDIS_MAKE_SHORTSTRING("cmpss"),
    ZYDIS_MAKE_SHORTSTRING("cmpsw"),
    ZYDIS_MAKE_SHORTSTRING("cmpxchg"),
    ZYDIS_MAKE_SHORTSTRING("cmpxchg16b"),
    ZYDIS_MAKE_SHORTSTRING("cmpxchg8b"),
    ZYDIS_MAKE_SHORTSTRING("comisd"),
    ZYDIS_MAKE_SHORTSTRING("comiss"),
    ZYDIS_MAKE_SHORTSTRING("cpuid"),
    ZYDIS_MAKE_SHORTSTRING("cqo"),
    ZYDIS_MAKE_SHORTSTRING("crc32"),
    ZYDIS_MAKE_SHORTSTRING("cvtdq2pd"),
    ZYDIS_MAKE_SHORTSTRING("cvtdq2ps"),
    ZYDIS_MAKE_SHORTSTRING("cvtpd2dq"),
    ZYDIS_MAKE_SHORTSTRING("cvtpd2pi"),
    ZYDIS_MAKE_SHORTSTRING("cvtpd2ps"),
    ZYDIS_MAKE_SHORTSTRING("cvtpi2pd"),
    ZYDIS_MAKE_SHORTSTRING("cvtpi2ps"),
    ZYDIS_MAKE_SHORTSTRING("cvtps2dq"),
    ZYDIS_MAKE_SHORTSTRING("cvtps2pd"),
    ZYDIS_MAKE_SHORTSTRING("cvtps2pi"),
    ZYDIS_MAKE_SHORTSTRING("cvtsd2si"),
    ZYDIS_MAKE_SHORTSTRING("cvtsd2ss"),
    ZYDIS_MAKE_SHORTSTRING("cvtsi2sd"),
    ZYDIS_MAKE_SHORTSTRING("cvtsi2ss"),
    ZYDIS_MAKE_SHORTSTRING("cvtss2sd"),
    ZYDIS_MAKE_SHORTSTRING("cvtss2si"),
    ZYDIS_MAKE_SHORTSTRING("cvttpd2dq"),
    ZYDIS_MAKE_SHORTSTRING("cvttpd2pi"),
    ZYDIS_MAKE_SHORTSTRING("cvttps2dq"),
    ZYDIS_MAKE_SHORTSTRING("cvttps2pi"),
    ZYDIS_MAKE_SHORTSTRING("cvttsd2si"),
    ZYDIS_MAKE_SHORTSTRING("cvttss2si"),
    ZYDIS_MAKE_SHORTSTRING("cwd"),
    ZYDIS_MAKE_SHORTSTRING("cwde"),
    ZYDIS_MAKE_SHORTSTRING("daa"),
    ZYDIS_MAKE_SHORTSTRING("das"),
    ZYDIS_MAKE_SHORTSTRING("dec"),
    ZYDIS_MAKE_SHORTSTRING("delay"),
    ZYDIS_MAKE_SHORTSTRING("div"),
    ZYDIS_MAKE_SHORTSTRING("divpd"),
    ZYDIS_MAKE_SHORTSTRING("divps"),
    ZYDIS_MAKE_SHORTSTRING("divsd"),
    ZYDIS_MAKE_SHORTSTRING("divss"),
    ZYDIS_MAKE_SHORTSTRING("dppd"),
    ZYDIS_MAKE_SHORTSTRING("dpps"),
    ZYDIS_MAKE_SHORTSTRING("emms"),
    ZYDIS_MAKE_SHORTSTRING("encls"),
    ZYDIS_MAKE_SHORTSTRING("enclu"),
    ZYDIS_MAKE_SHORTSTRING("enclv"),
    ZYDIS_MAKE_SHORTSTRING("encodekey128"),
    ZYDIS_MAKE_SHORTSTRING("encodekey256"),
    ZYDIS_MAKE_SHORTSTRING("endbr32"),
    ZYDIS_MAKE_SHORTSTRING("endbr64"),
    ZYDIS_MAKE_SHORTSTRING("enqcmd"),
    ZYDIS_MAKE_SHORTSTRING("enqcmds"),
    ZYDIS_MAKE_SHORTSTRING("enter"),
    ZYDIS_MAKE_SHORTSTRING("extractps"),
    ZYDIS_MAKE_SHORTSTRING("extrq"),
    ZYDIS_MAKE_SHORTSTRING("f2xm1"),
    ZYDIS_MAKE_SHORTSTRING("fabs"),
    ZYDIS_MAKE_SHORTSTRING("fadd"),
    ZYDIS_MAKE_SHORTSTRING("faddp"),
    ZYDIS_MAKE_SHORTSTRING("fbld"),
    ZYDIS_MAKE_SHORTSTRING("fbstp"),
    ZYDIS_MAKE_SHORTSTRING("fchs"),
    ZYDIS_MAKE_SHORTSTRING("fcmovb"),
    ZYDIS_MAKE_SHORTSTRING("fcmovbe"),
    ZYDIS_MAKE_SHORTSTRING("fcmove"),
    ZYDIS_MAKE_SHORTSTRING("fcmovnb"),
    ZYDIS_MAKE_SHORTSTRING("fcmovnbe"),
    ZYDIS_MAKE_SHORTSTRING("fcmovne"),
    ZYDIS_MAKE_SHORTSTRING("fcmovnu"),
    ZYDIS_MAKE_SHORTSTRING("fcmovu"),
    ZYDIS_MAKE_SHORTSTRING("fcom"),
    ZYDIS_MAKE_SHORTSTRING("fcomi"),
    ZYDIS_MAKE_SHORTSTRING("fcomip"),
    ZYDIS_MAKE_SHORTSTRING("fcomp"),
    ZYDIS_MAKE_SHORTSTRING("fcompp"),
    ZYDIS_MAKE_SHORTSTRING("fcos"),
    ZYDIS_MAKE_SHORTSTRING("fdecstp"),
    ZYDIS_MAKE_SHORTSTRING("fdisi8087_nop"),
    ZYDIS_MAKE_SHORTSTRING("fdiv"),
    ZYDIS_MAKE_SHORTSTRING("fdivp"),
    ZYDIS_MAKE_SHORTSTRING("fdivr"),
    ZYDIS_MAKE_SHORTSTRING("fdivrp"),
    ZYDIS_MAKE_SHORTSTRING("femms"),
    ZYDIS_MAKE_SHORTSTRING("feni8087_nop"),
    ZYDIS_MAKE_SHORTSTRING("ffree"),
    ZYDIS_MAKE_SHORTSTRING("ffreep"),
    ZYDIS_MAKE_SHORTSTRING("fiadd"),
    ZYDIS_MAKE_SHORTSTRING("ficom"),
    ZYDIS_MAKE_SHORTSTRING("ficomp"),
    ZYDIS_MAKE_SHORTSTRING("fidiv"),
    ZYDIS_MAKE_SHORTSTRING("fidivr"),
    ZYDIS_MAKE_SHORTSTRING("fild"),
    ZYDIS_MAKE_SHORTSTRING("fimul"),
    ZYDIS_MAKE_SHORTSTRING("fincstp"),
    ZYDIS_MAKE_SHORTSTRING("fist"),
    ZYDIS_MAKE_SHORTSTRING("fistp"),
    ZYDIS_MAKE_SHORTSTRING("fisttp"),
    ZYDIS_MAKE_SHORTSTRING("fisub"),
    ZYDIS_MAKE_SHORTSTRING("fisubr"),
    ZYDIS_MAKE_SHORTSTRING("fld"),
    ZYDIS_MAKE_SHORTSTRING("fld1"),
    ZYDIS_MAKE_SHORTSTRING("fldcw"),
    ZYDIS_MAKE_SHORTSTRING("fldenv"),
    ZYDIS_MAKE_SHORTSTRING("fldl2e"),
    ZYDIS_MAKE_SHORTSTRING("fldl2t"),
    ZYDIS_MAKE_SHORTSTRING("fldlg2"),
    ZYDIS_MAKE_SHORTSTRING("fldln2"),
    ZYDIS_MAKE_SHORTSTRING("fldpi"),
    ZYDIS_MAKE_SHORTSTRING("fldz"),
    ZYDIS_MAKE_SHORTSTRING("fmul"),
    ZYDIS_MAKE_SHORTSTRING("fmulp"),
    ZYDIS_MAKE_SHORTSTRING("fnclex"),
    ZYDIS_MAKE_SHORTSTRING("fninit"),
    ZYDIS_MAKE_SHORTSTRING("fnop"),
    ZYDIS_MAKE_SHORTSTRING("fnsave"),
    ZYDIS_MAKE_SHORTSTRING("fnstcw"),
    ZYDIS_MAKE_SHORTSTRING("fnstenv"),
    ZYDIS_MAKE_SHORTSTRING("fnstsw"),
    ZYDIS_MAKE_SHORTSTRING("fpatan"),
    ZYDIS_MAKE_SHORTSTRING("fprem"),
    ZYDIS_MAKE_SHORTSTRING("fprem1"),
    ZYDIS_MAKE_SHORTSTRING("fptan"),
    ZYDIS_MAKE_SHORTSTRING("frndint"),
    ZYDIS_MAKE_SHORTSTRING("frstor"),
    ZYDIS_MAKE_SHORTSTRING("fscale"),
    ZYDIS_MAKE_SHORTSTRING("fsetpm287_nop"),
    ZYDIS_MAKE_SHORTSTRING("fsin"),
    ZYDIS_MAKE_SHORTSTRING("fsincos"),
    ZYDIS_MAKE_SHORTSTRING("fsqrt"),
    ZYDIS_MAKE_SHORTSTRING("fst"),
    ZYDIS_MAKE_SHORTSTRING("fstp"),
    ZYDIS_MAKE_SHORTSTRING("fstpnce"),
    ZYDIS_MAKE_SHORTSTRING("fsub"),
    ZYDIS_MAKE_SHORTSTRING("fsubp"),
    ZYDIS_MAKE_SHORTSTRING("fsubr"),
    ZYDIS_MAKE_SHORTSTRING("fsubrp"),
    ZYDIS_MAKE_SHORTSTRING("ftst"),
    ZYDIS_MAKE_SHORTSTRING("fucom"),
    ZYDIS_MAKE_SHORTSTRING("fucomi"),
    ZYDIS_MAKE_SHORTSTRING("fucomip"),
    ZYDIS_MAKE_SHORTSTRING("fucomp"),
    ZYDIS_MAKE_SHORTSTRING("fucompp"),
    ZYDIS_MAKE_SHORTSTRING("fwait"),
    ZYDIS_MAKE_SHORTSTRING("fxam"),
    ZYDIS_MAKE_SHORTSTRING("fxch"),
    ZYDIS_MAKE_SHORTSTRING("fxrstor"),
    ZYDIS_MAKE_SHORTSTRING("fxrstor64"),
    ZYDIS_MAKE_SHORTSTRING("fxsave"),
    ZYDIS_MAKE_SHORTSTRING("fxsave64"),
    ZYDIS_MAKE_SHORTSTRING("fxtract"),
    ZYDIS_MAKE_SHORTSTRING("fyl2x"),
    ZYDIS_MAKE_SHORTSTRING("fyl2xp1"),
    ZYDIS_MAKE_SHORTSTRING("getsec"),
    ZYDIS_MAKE_SHORTSTRING("gf2p8affineinvqb"),
    ZYDIS_MAKE_SHORTSTRING("gf2p8affineqb"),
    ZYDIS_MAKE_SHORTSTRING("gf2p8mulb"),
    ZYDIS_MAKE_SHORTSTRING("haddpd"),
    ZYDIS_MAKE_SHORTSTRING("haddps"),
    ZYDIS_MAKE_SHORTSTRING("hlt"),
    ZYDIS_MAKE_SHORTSTRING("hreset"),
    ZYDIS_MAKE_SHORTSTRING("hsubpd"),
    ZYDIS_MAKE_SHORTSTRING("hsubps"),
    ZYDIS_MAKE_SHORTSTRING("idiv"),
    ZYDIS_MAKE_SHORTSTRING("imul"),
    ZYDIS_MAKE_SHORTSTRING("in"),
    ZYDIS_MAKE_SHORTSTRING("inc"),
    ZYDIS_MAKE_SHORTSTRING("incsspd"),
    ZYDIS_MAKE_SHORTSTRING("incsspq"),
    ZYDIS_MAKE_SHORTSTRING("insb"),
    ZYDIS_MAKE_SHORTSTRING("insd"),
    ZYDIS_MAKE_SHORTSTRING("insertps"),
    ZYDIS_MAKE_SHORTSTRING("insertq"),
    ZYDIS_MAKE_SHORTSTRING("insw"),
    ZYDIS_MAKE_SHORTSTRING("int"),
    ZYDIS_MAKE_SHORTSTRING("int1"),
    ZYDIS_MAKE_SHORTSTRING("int3"),
    ZYDIS_MAKE_SHORTSTRING("into"),
    ZYDIS_MAKE_SHORTSTRING("invd"),
    ZYDIS_MAKE_SHORTSTRING("invept"),
    ZYDIS_MAKE_SHORTSTRING("invlpg"),
    ZYDIS_MAKE_SHORTSTRING("invlpga"),
    ZYDIS_MAKE_SHORTSTRING("invlpgb"),
    ZYDIS_MAKE_SHORTSTRING("invpcid"),
    ZYDIS_MAKE_SHORTSTRING("invvpid"),
    ZYDIS_MAKE_SHORTSTRING("iret"),
    ZYDIS_MAKE_SHORTSTRING("iretd"),
    ZYDIS_MAKE_SHORTSTRING("iretq"),
    ZYDIS_MAKE_SHORTSTRING("jb"),
    ZYDIS_MAKE_SHORTSTRING("jbe"),
    ZYDIS_MAKE_SHORTSTRING("jcxz"),
    ZYDIS_MAKE_SHORTSTRING("jecxz"),
    ZYDIS_MAKE_SHORTSTRING("jknzd"),
    ZYDIS_MAKE_SHORTSTRING("jkzd"),
    ZYDIS_MAKE_SHORTSTRING("jl"),
    ZYDIS_MAKE_SHORTSTRING("jle"),
    ZYDIS_MAKE_SHORTSTRING("jmp"),
    ZYDIS_MAKE_SHORTSTRING("jnb"),
    ZYDIS_MAKE_SHORTSTRING("jnbe"),
    ZYDIS_MAKE_SHORTSTRING("jnl"),
    ZYDIS_MAKE_SHORTSTRING("jnle"),
    ZYDIS_MAKE_SHORTSTRING("jno"),
    ZYDIS_MAKE_SHORTSTRING("jnp"),
    ZYDIS_MAKE_SHORTSTRING("jns"),
    ZYDIS_MAKE_SHORTSTRING("jnz"),
    ZYDIS_MAKE_SHORTSTRING("jo"),
    ZYDIS_MAKE_SHORTSTRING("jp"),
    ZYDIS_MAKE_SHORTSTRING("jrcxz"),
    ZYDIS_MAKE_SHORTSTRING("js"),
    ZYDIS_MAKE_SHORTSTRING("jz"),
    ZYDIS_MAKE_SHORTSTRING("kaddb"),
    ZYDIS_MAKE_SHORTSTRING("kaddd"),
    ZYDIS_MAKE_SHORTSTRING("kaddq"),
    ZYDIS_MAKE_SHORTSTRING("kaddw"),
    ZYDIS_MAKE_SHORTSTRING("kand"),
    ZYDIS_MAKE_SHORTSTRING("kandb"),
    ZYDIS_MAKE_SHORTSTRING("kandd"),
    ZYDIS_MAKE_SHORTSTRING("kandn"),
    ZYDIS_MAKE_SHORTSTRING("kandnb"),
    ZYDIS_MAKE_SHORTSTRING("kandnd"),
    ZYDIS_MAKE_SHORTSTRING("kandnq"),
    ZYDIS_MAKE_SHORTSTRING("kandnr"),
    ZYDIS_MAKE_SHORTSTRING("kandnw"),
    ZYDIS_MAKE_SHORTSTRING("kandq"),
    ZYDIS_MAKE_SHORTSTRING("kandw"),
    ZYDIS_MAKE_SHORTSTRING("kconcath"),
    ZYDIS_MAKE_SHORTSTRING("kconcatl"),
    ZYDIS_MAKE_SHORTSTRING("kextract"),
    ZYDIS_MAKE_SHORTSTRING("kmerge2l1h"),
    ZYDIS_MAKE_SHORTSTRING("kmerge2l1l"),
    ZYDIS_MAKE_SHORTSTRING("kmov"),
    ZYDIS_MAKE_SHORTSTRING("kmovb"),
    ZYDIS_MAKE_SHORTSTRING("kmovd"),
    ZYDIS_MAKE_SHORTSTRING("kmovq"),
    ZYDIS_MAKE_SHORTSTRING("kmovw"),
    ZYDIS_MAKE_SHORTSTRING("knot"),
    ZYDIS_MAKE_SHORTSTRING("knotb"),
    ZYDIS_MAKE_SHORTSTRING("knotd"),
    ZYDIS_MAKE_SHORTSTRING("knotq"),
    ZYDIS_MAKE_SHORTSTRING("knotw"),
    ZYDIS_MAKE_SHORTSTRING("kor"),
    ZYDIS_MAKE_SHORTSTRING("korb"),
    ZYDIS_MAKE_SHORTSTRING("kord"),
    ZYDIS_MAKE_SHORTSTRING("korq"),
    ZYDIS_MAKE_SHORTSTRING("kortest"),
    ZYDIS_MAKE_SHORTSTRING("kortestb"),
    ZYDIS_MAKE_SHORTSTRING("kortestd"),
    ZYDIS_MAKE_SHORTSTRING("kortestq"),
    ZYDIS_MAKE_SHORTSTRING("kortestw"),
    ZYDIS_MAKE_SHORTSTRING("korw"),
    ZYDIS_MAKE_SHORTSTRING("kshiftlb"),
    ZYDIS_MAKE_SHORTSTRING("kshiftld"),
    ZYDIS_MAKE_SHORTSTRING("kshiftlq"),
    ZYDIS_MAKE_SHORTSTRING("kshiftlw"),
    ZYDIS_MAKE_SHORTSTRING("kshiftrb"),
    ZYDIS_MAKE_SHORTSTRING("kshiftrd"),
    ZYDIS_MAKE_SHORTSTRING("kshiftrq"),
    ZYDIS_MAKE_SHORTSTRING("kshiftrw"),
    ZYDIS_MAKE_SHORTSTRING("ktestb"),
    ZYDIS_MAKE_SHORTSTRING("ktestd"),
    ZYDIS_MAKE_SHORTSTRING("ktestq"),
    ZYDIS_MAKE_SHORTSTRING("ktestw"),
    ZYDIS_MAKE_SHORTSTRING("kunpckbw"),
    ZYDIS_MAKE_SHORTSTRING("kunpckdq"),
    ZYDIS_MAKE_SHORTSTRING("kunpckwd"),
    ZYDIS_MAKE_SHORTSTRING("kxnor"),
    ZYDIS_MAKE_SHORTSTRING("kxnorb"),
    ZYDIS_MAKE_SHORTSTRING("kxnord"),
    ZYDIS_MAKE_SHORTSTRING("kxnorq"),
    ZYDIS_MAKE_SHORTSTRING("kxnorw"),
    ZYDIS_MAKE_SHORTSTRING("kxor"),
    ZYDIS_MAKE_SHORTSTRING("kxorb"),
    ZYDIS_MAKE_SHORTSTRING("kxord"),
    ZYDIS_MAKE_SHORTSTRING("kxorq"),
    ZYDIS_MAKE_SHORTSTRING("kxorw"),
    ZYDIS_MAKE_SHORTSTRING("lahf"),
    ZYDIS_MAKE_SHORTSTRING("lar"),
    ZYDIS_MAKE_SHORTSTRING("lddqu"),
    ZYDIS_MAKE_SHORTSTRING("ldmxcsr"),
    ZYDIS_MAKE_SHORTSTRING("lds"),
    ZYDIS_MAKE_SHORTSTRING("ldtilecfg"),
    ZYDIS_MAKE_SHORTSTRING("lea"),
    ZYDIS_MAKE_SHORTSTRING("leave"),
    ZYDIS_MAKE_SHORTSTRING("les"),
    ZYDIS_MAKE_SHORTSTRING("lfence"),
    ZYDIS_MAKE_SHORTSTRING("lfs"),
    ZYDIS_MAKE_SHORTSTRING("lgdt"),
    ZYDIS_MAKE_SHORTSTRING("lgs"),
    ZYDIS_MAKE_SHORTSTRING("lidt"),
    ZYDIS_MAKE_SHORTSTRING("lldt"),
    ZYDIS_MAKE_SHORTSTRING("llwpcb"),
    ZYDIS_MAKE_SHORTSTRING("lmsw"),
    ZYDIS_MAKE_SHORTSTRING("loadiwkey"),
    ZYDIS_MAKE_SHORTSTRING("lodsb"),
    ZYDIS_MAKE_SHORTSTRING("lodsd"),
    ZYDIS_MAKE_SHORTSTRING("lodsq"),
    ZYDIS_MAKE_SHORTSTRING("lodsw"),
    ZYDIS_MAKE_SHORTSTRING("loop"),
    ZYDIS_MAKE_SHORTSTRING("loope"),
    ZYDIS_MAKE_SHORTSTRING("loopne"),
    ZYDIS_MAKE_SHORTSTRING("lsl"),
    ZYDIS_MAKE_SHORTSTRING("lss"),
    ZYDIS_MAKE_SHORTSTRING("ltr"),
    ZYDIS_MAKE_SHORTSTRING("lwpins"),
    ZYDIS_MAKE_SHORTSTRING("lwpval"),
    ZYDIS_MAKE_SHORTSTRING("lzcnt"),
    ZYDIS_MAKE_SHORTSTRING("maskmovdqu"),
    ZYDIS_MAKE_SHORTSTRING("maskmovq"),
    ZYDIS_MAKE_SHORTSTRING("maxpd"),
    ZYDIS_MAKE_SHORTSTRING("maxps"),
    ZYDIS_MAKE_SHORTSTRING("maxsd"),
    ZYDIS_MAKE_SHORTSTRING("maxss"),
    ZYDIS_MAKE_SHORTSTRING("mcommit"),
    ZYDIS_MAKE_SHORTSTRING("mfence"),
    ZYDIS_MAKE_SHORTSTRING("minpd"),
    ZYDIS_MAKE_SHORTSTRING("minps"),
    ZYDIS_MAKE_SHORTSTRING("minsd"),
    ZYDIS_MAKE_SHORTSTRING("minss"),
    ZYDIS_MAKE_SHORTSTRING("monitor"),
    ZYDIS_MAKE_SHORTSTRING("monitorx"),
    ZYDIS_MAKE_SHORTSTRING("montmul"),
    ZYDIS_MAKE_SHORTSTRING("mov"),
    ZYDIS_MAKE_SHORTSTRING("movapd"),
    ZYDIS_MAKE_SHORTSTRING("movaps"),
    ZYDIS_MAKE_SHORTSTRING("movbe"),
    ZYDIS_MAKE_SHORTSTRING("movd"),
    ZYDIS_MAKE_SHORTSTRING("movddup"),
    ZYDIS_MAKE_SHORTSTRING("movdir64b"),
    ZYDIS_MAKE_SHORTSTRING("movdiri"),
    ZYDIS_MAKE_SHORTSTRING("movdq2q"),
    ZYDIS_MAKE_SHORTSTRING("movdqa"),
    ZYDIS_MAKE_SHORTSTRING("movdqu"),
    ZYDIS_MAKE_SHORTSTRING("movhlps"),
    ZYDIS_MAKE_SHORTSTRING("movhpd"),
    ZYDIS_MAKE_SHORTSTRING("movhps"),
    ZYDIS_MAKE_SHORTSTRING("movlhps"),
    ZYDIS_MAKE_SHORTSTRING("movlpd"),
    ZYDIS_MAKE_SHORTSTRING("movlps"),
    ZYDIS_MAKE_SHORTSTRING("movmskpd"),
    ZYDIS_MAKE_SHORTSTRING("movmskps"),
    ZYDIS_MAKE_SHORTSTRING("movntdq"),
    ZYDIS_MAKE_SHORTSTRING("movntdqa"),
    ZYDIS_MAKE_SHORTSTRING("movnti"),
    ZYDIS_MAKE_SHORTSTRING("movntpd"),
    ZYDIS_MAKE_SHORTSTRING("movntps"),
    ZYDIS_MAKE_SHORTSTRING("movntq"),
    ZYDIS_MAKE_SHORTSTRING("movntsd"),
    ZYDIS_MAKE_SHORTSTRING("movntss"),
    ZYDIS_MAKE_SHORTSTRING("movq"),
    ZYDIS_MAKE_SHORTSTRING("movq2dq"),
    ZYDIS_MAKE_SHORTSTRING("movsb"),
    ZYDIS_MAKE_SHORTSTRING("movsd"),
    ZYDIS_MAKE_SHORTSTRING("movshdup"),
    ZYDIS_MAKE_SHORTSTRING("movsldup"),
    ZYDIS_MAKE_SHORTSTRING("movsq"),
    ZYDIS_MAKE_SHORTSTRING("movss"),
    ZYDIS_MAKE_SHORTSTRING("movsw"),
    ZYDIS_MAKE_SHORTSTRING("movsx"),
    ZYDIS_MAKE_SHORTSTRING("movsxd"),
    ZYDIS_MAKE_SHORTSTRING("movupd"),
    ZYDIS_MAKE_SHORTSTRING("movups"),
    ZYDIS_MAKE_SHORTSTRING("movzx"),
    ZYDIS_MAKE_SHORTSTRING("mpsadbw"),
    ZYDIS_MAKE_SHORTSTRING("mul"),
    ZYDIS_MAKE_SHORTSTRING("mulpd"),
    ZYDIS_MAKE_SHORTSTRING("mulps"),
    ZYDIS_MAKE_SHORTSTRING("mulsd"),
    ZYDIS_MAKE_SHORTSTRING("mulss"),
    ZYDIS_MAKE_SHORTSTRING("mulx"),
    ZYDIS_MAKE_SHORTSTRING("mwait"),
    ZYDIS_MAKE_SHORTSTRING("mwaitx"),
    ZYDIS_MAKE_SHORTSTRING("neg"),
    ZYDIS_MAKE_SHORTSTRING("nop"),
    ZYDIS_MAKE_SHORTSTRING("not"),
    ZYDIS_MAKE_SHORTSTRING("or"),
    ZYDIS_MAKE_SHORTSTRING("orpd"),
    ZYDIS_MAKE_SHORTSTRING("orps"),
    ZYDIS_MAKE_SHORTSTRING("out"),
    ZYDIS_MAKE_SHORTSTRING("outsb"),
    ZYDIS_MAKE_SHORTSTRING("outsd"),
    ZYDIS_MAKE_SHORTSTRING("outsw"),
    ZYDIS_MAKE_SHORTSTRING("pabsb"),
    ZYDIS_MAKE_SHORTSTRING("pabsd"),
    ZYDIS_MAKE_SHORTSTRING("pabsw"),
    ZYDIS_MAKE_SHORTSTRING("packssdw"),
    ZYDIS_MAKE_SHORTSTRING("packsswb"),
    ZYDIS_MAKE_SHORTSTRING("packusdw"),
    ZYDIS_MAKE_SHORTSTRING("packuswb"),
    ZYDIS_MAKE_SHORTSTRING("paddb"),
    ZYDIS_MAKE_SHORTSTRING("paddd"),
    ZYDIS_MAKE_SHORTSTRING("paddq"),
    ZYDIS_MAKE_SHORTSTRING("paddsb"),
    ZYDIS_MAKE_SHORTSTRING("paddsw"),
    ZYDIS_MAKE_SHORTSTRING("paddusb"),
    ZYDIS_MAKE_SHORTSTRING("paddusw"),
    ZYDIS_MAKE_SHORTSTRING("paddw"),
    ZYDIS_MAKE_SHORTSTRING("palignr"),
    ZYDIS_MAKE_SHORTSTRING("pand"),
    ZYDIS_MAKE_SHORTSTRING("pandn"),
    ZYDIS_MAKE_SHORTSTRING("pause"),
    ZYDIS_MAKE_SHORTSTRING("pavgb"),
    ZYDIS_MAKE_SHORTSTRING("pavgusb"),
    ZYDIS_MAKE_SHORTSTRING("pavgw"),
    ZYDIS_MAKE_SHORTSTRING("pblendvb"),
    ZYDIS_MAKE_SHORTSTRING("pblendw"),
    ZYDIS_MAKE_SHORTSTRING("pclmulqdq"),
    ZYDIS_MAKE_SHORTSTRING("pcmpeqb"),
    ZYDIS_MAKE_SHORTSTRING("pcmpeqd"),
    ZYDIS_MAKE_SHORTSTRING("pcmpeqq"),
    ZYDIS_MAKE_SHORTSTRING("pcmpeqw"),
    ZYDIS_MAKE_SHORTSTRING("pcmpestri"),
    ZYDIS_MAKE_SHORTSTRING("pcmpestrm"),
    ZYDIS_MAKE_SHORTSTRING("pcmpgtb"),
    ZYDIS_MAKE_SHORTSTRING("pcmpgtd"),
    ZYDIS_MAKE_SHORTSTRING("pcmpgtq"),
    ZYDIS_MAKE_SHORTSTRING("pcmpgtw"),
    ZYDIS_MAKE_SHORTSTRING("pcmpistri"),
    ZYDIS_MAKE_SHORTSTRING("pcmpistrm"),
    ZYDIS_MAKE_SHORTSTRING("pconfig"),
    ZYDIS_MAKE_SHORTSTRING("pdep"),
    ZYDIS_MAKE_SHORTSTRING("pext"),
    ZYDIS_MAKE_SHORTSTRING("pextrb"),
    ZYDIS_MAKE_SHORTSTRING("pextrd"),
    ZYDIS_MAKE_SHORTSTRING("pextrq"),
    ZYDIS_MAKE_SHORTSTRING("pextrw"),
    ZYDIS_MAKE_SHORTSTRING("pf2id"),
    ZYDIS_MAKE_SHORTSTRING("pf2iw"),
    ZYDIS_MAKE_SHORTSTRING("pfacc"),
    ZYDIS_MAKE_SHORTSTRING("pfadd"),
    ZYDIS_MAKE_SHORTSTRING("pfcmpeq"),
    ZYDIS_MAKE_SHORTSTRING("pfcmpge"),
    ZYDIS_MAKE_SHORTSTRING("pfcmpgt"),
    ZYDIS_MAKE_SHORTSTRING("pfcpit1"),
    ZYDIS_MAKE_SHORTSTRING("pfmax"),
    ZYDIS_MAKE_SHORTSTRING("pfmin"),
    ZYDIS_MAKE_SHORTSTRING("pfmul"),
    ZYDIS_MAKE_SHORTSTRING("pfnacc"),
    ZYDIS_MAKE_SHORTSTRING("pfpnacc"),
    ZYDIS_MAKE_SHORTSTRING("pfrcp"),
    ZYDIS_MAKE_SHORTSTRING("pfrcpit2"),
    ZYDIS_MAKE_SHORTSTRING("pfrsqit1"),
    ZYDIS_MAKE_SHORTSTRING("pfsqrt"),
    ZYDIS_MAKE_SHORTSTRING("pfsub"),
    ZYDIS_MAKE_SHORTSTRING("pfsubr"),
    ZYDIS_MAKE_SHORTSTRING("phaddd"),
    ZYDIS_MAKE_SHORTSTRING("phaddsw"),
    ZYDIS_MAKE_SHORTSTRING("phaddw"),
    ZYDIS_MAKE_SHORTSTRING("phminposuw"),
    ZYDIS_MAKE_SHORTSTRING("phsubd"),
    ZYDIS_MAKE_SHORTSTRING("phsubsw"),
    ZYDIS_MAKE_SHORTSTRING("phsubw"),
    ZYDIS_MAKE_SHORTSTRING("pi2fd"),
    ZYDIS_MAKE_SHORTSTRING("pi2fw"),
    ZYDIS_MAKE_SHORTSTRING("pinsrb"),
    ZYDIS_MAKE_SHORTSTRING("pinsrd"),
    ZYDIS_MAKE_SHORTSTRING("pinsrq"),
    ZYDIS_MAKE_SHORTSTRING("pinsrw"),
    ZYDIS_MAKE_SHORTSTRING("pmaddubsw"),
    ZYDIS_MAKE_SHORTSTRING("pmaddwd"),
    ZYDIS_MAKE_SHORTSTRING("pmaxsb"),
    ZYDIS_MAKE_SHORTSTRING("pmaxsd"),
    ZYDIS_MAKE_SHORTSTRING("pmaxsw"),
    ZYDIS_MAKE_SHORTSTRING("pmaxub"),
    ZYDIS_MAKE_SHORTSTRING("pmaxud"),
    ZYDIS_MAKE_SHORTSTRING("pmaxuw"),
    ZYDIS_MAKE_SHORTSTRING("pminsb"),
    ZYDIS_MAKE_SHORTSTRING("pminsd"),
    ZYDIS_MAKE_SHORTSTRING("pminsw"),
    ZYDIS_MAKE_SHORTSTRING("pminub"),
    ZYDIS_MAKE_SHORTSTRING("pminud"),
    ZYDIS_MAKE_SHORTSTRING("pminuw"),
    ZYDIS_MAKE_SHORTSTRING("pmovmskb"),
    ZYDIS_MAKE_SHORTSTRING("pmovsxbd"),
    ZYDIS_MAKE_SHORTSTRING("pmovsxbq"),
    ZYDIS_MAKE_SHORTSTRING("pmovsxbw"),
    ZYDIS_MAKE_SHORTSTRING("pmovsxdq"),
    ZYDIS_MAKE_SHORTSTRING("pmovsxwd"),
    ZYDIS_MAKE_SHORTSTRING("pmovsxwq"),
    ZYDIS_MAKE_SHORTSTRING("pmovzxbd"),
    ZYDIS_MAKE_SHORTSTRING("pmovzxbq"),
    ZYDIS_MAKE_SHORTSTRING("pmovzxbw"),
    ZYDIS_MAKE_SHORTSTRING("pmovzxdq"),
    ZYDIS_MAKE_SHORTSTRING("pmovzxwd"),
    ZYDIS_MAKE_SHORTSTRING("pmovzxwq"),
    ZYDIS_MAKE_SHORTSTRING("pmuldq"),
    ZYDIS_MAKE_SHORTSTRING("pmulhrsw"),
    ZYDIS_MAKE_SHORTSTRING("pmulhrw"),
    ZYDIS_MAKE_SHORTSTRING("pmulhuw"),
    ZYDIS_MAKE_SHORTSTRING("pmulhw"),
    ZYDIS_MAKE_SHORTSTRING("pmulld"),
    ZYDIS_MAKE_SHORTSTRING("pmullw"),
    ZYDIS_MAKE_SHORTSTRING("pmuludq"),
    ZYDIS_MAKE_SHORTSTRING("pop"),
    ZYDIS_MAKE_SHORTSTRING("popa"),
    ZYDIS_MAKE_SHORTSTRING("popad"),
    ZYDIS_MAKE_SHORTSTRING("popcnt"),
    ZYDIS_MAKE_SHORTSTRING("popf"),
    ZYDIS_MAKE_SHORTSTRING("popfd"),
    ZYDIS_MAKE_SHORTSTRING("popfq"),
    ZYDIS_MAKE_SHORTSTRING("por"),
    ZYDIS_MAKE_SHORTSTRING("prefetch"),
    ZYDIS_MAKE_SHORTSTRING("prefetchnta"),
    ZYDIS_MAKE_SHORTSTRING("prefetcht0"),
    ZYDIS_MAKE_SHORTSTRING("prefetcht1"),
    ZYDIS_MAKE_SHORTSTRING("prefetcht2"),
    ZYDIS_MAKE_SHORTSTRING("prefetchw"),
    ZYDIS_MAKE_SHORTSTRING("prefetchwt1"),
    ZYDIS_MAKE_SHORTSTRING("psadbw"),
    ZYDIS_MAKE_SHORTSTRING("pshufb"),
    ZYDIS_MAKE_SHORTSTRING("pshufd"),
    ZYDIS_MAKE_SHORTSTRING("pshufhw"),
    ZYDIS_MAKE_SHORTSTRING("pshuflw"),
    ZYDIS_MAKE_SHORTSTRING("pshufw"),
    ZYDIS_MAKE_SHORTSTRING("psignb"),
    ZYDIS_MAKE_SHORTSTRING("psignd"),
    ZYDIS_MAKE_SHORTSTRING("psignw"),
    ZYDIS_MAKE_SHORTSTRING("pslld"),
    ZYDIS_MAKE_SHORTSTRING("pslldq"),
    ZYDIS_MAKE_SHORTSTRING("psllq"),
    ZYDIS_MAKE_SHORTSTRING("psllw"),
    ZYDIS_MAKE_SHORTSTRING("psmash"),
    ZYDIS_MAKE_SHORTSTRING("psrad"),
    ZYDIS_MAKE_SHORTSTRING("psraw"),
    ZYDIS_MAKE_SHORTSTRING("psrld"),
    ZYDIS_MAKE_SHORTSTRING("psrldq"),
    ZYDIS_MAKE_SHORTSTRING("psrlq"),
    ZYDIS_MAKE_SHORTSTRING("psrlw"),
    ZYDIS_MAKE_SHORTSTRING("psubb"),
    ZYDIS_MAKE_SHORTSTRING("psubd"),
    ZYDIS_MAKE_SHORTSTRING("psubq"),
    ZYDIS_MAKE_SHORTSTRING("psubsb"),
    ZYDIS_MAKE_SHORTSTRING("psubsw"),
    ZYDIS_MAKE_SHORTSTRING("psubusb"),
    ZYDIS_MAKE_SHORTSTRING("psubusw"),
    ZYDIS_MAKE_SHORTSTRING("psubw"),
    ZYDIS_MAKE_SHORTSTRING("pswapd"),
    ZYDIS_MAKE_SHORTSTRING("ptest"),
    ZYDIS_MAKE_SHORTSTRING("ptwrite"),
    ZYDIS_MAKE_SHORTSTRING("punpckhbw"),
    ZYDIS_MAKE_SHORTSTRING("punpckhdq"),
    ZYDIS_MAKE_SHORTSTRING("punpckhqdq"),
    ZYDIS_MAKE_SHORTSTRING("punpckhwd"),
    ZYDIS_MAKE_SHORTSTRING("punpcklbw"),
    ZYDIS_MAKE_SHORTSTRING("punpckldq"),
    ZYDIS_MAKE_SHORTSTRING("punpcklqdq"),
    ZYDIS_MAKE_SHORTSTRING("punpcklwd"),
    ZYDIS_MAKE_SHORTSTRING("push"),
    ZYDIS_MAKE_SHORTSTRING("pusha"),
    ZYDIS_MAKE_SHORTSTRING("pushad"),
    ZYDIS_MAKE_SHORTSTRING("pushf"),
    ZYDIS_MAKE_SHORTSTRING("pushfd"),
    ZYDIS_MAKE_SHORTSTRING("pushfq"),
    ZYDIS_MAKE_SHORTSTRING("pvalidate"),
    ZYDIS_MAKE_SHORTSTRING("pxor"),
    ZYDIS_MAKE_SHORTSTRING("rcl"),
    ZYDIS_MAKE_SHORTSTRING("rcpps"),
    ZYDIS_MAKE_SHORTSTRING("rcpss"),
    ZYDIS_MAKE_SHORTSTRING("rcr"),
    ZYDIS_MAKE_SHORTSTRING("rdfsbase"),
    ZYDIS_MAKE_SHORTSTRING("rdgsbase"),
    ZYDIS_MAKE_SHORTSTRING("rdmsr"),
    ZYDIS_MAKE_SHORTSTRING("rdpid"),
    ZYDIS_MAKE_SHORTSTRING("rdpkru"),
    ZYDIS_MAKE_SHORTSTRING("rdpmc"),
    ZYDIS_MAKE_SHORTSTRING("rdpru"),
    ZYDIS_MAKE_SHORTSTRING("rdrand"),
    ZYDIS_MAKE_SHORTSTRING("rdseed"),
    ZYDIS_MAKE_SHORTSTRING("rdsspd"),
    ZYDIS_MAKE_SHORTSTRING("rdsspq"),
    ZYDIS_MAKE_SHORTSTRING("rdtsc"),
    ZYDIS_MAKE_SHORTSTRING("rdtscp"),
    ZYDIS_MAKE_SHORTSTRING("ret"),
    ZYDIS_MAKE_SHORTSTRING("rmpadjust"),
    ZYDIS_MAKE_SHORTSTRING("rmpupdate"),
    ZYDIS_MAKE_SHORTSTRING("rol"),
    ZYDIS_MAKE_SHORTSTRING("ror"),
    ZYDIS_MAKE_SHORTSTRING("rorx"),
    ZYDIS_MAKE_SHORTSTRING("roundpd"),
    ZYDIS_MAKE_SHORTSTRING("roundps"),
    ZYDIS_MAKE_SHORTSTRING("roundsd"),
    ZYDIS_MAKE_SHORTSTRING("roundss"),
    ZYDIS_MAKE_SHORTSTRING("rsm"),
    ZYDIS_MAKE_SHORTSTRING("rsqrtps"),
    ZYDIS_MAKE_SHORTSTRING("rsqrtss"),
    ZYDIS_MAKE_SHORTSTRING("rstorssp"),
    ZYDIS_MAKE_SHORTSTRING("sahf"),
    ZYDIS_MAKE_SHORTSTRING("salc"),
    ZYDIS_MAKE_SHORTSTRING("sar"),
    ZYDIS_MAKE_SHORTSTRING("sarx"),
    ZYDIS_MAKE_SHORTSTRING("saveprevssp"),
    ZYDIS_MAKE_SHORTSTRING("sbb"),
    ZYDIS_MAKE_SHORTSTRING("scasb"),
    ZYDIS_MAKE_SHORTSTRING("scasd"),
    ZYDIS_MAKE_SHORTSTRING("scasq"),
    ZYDIS_MAKE_SHORTSTRING("scasw"),
    ZYDIS_MAKE_SHORTSTRING("seamcall"),
    ZYDIS_MAKE_SHORTSTRING("seamops"),
    ZYDIS_MAKE_SHORTSTRING("seamret"),
    ZYDIS_MAKE_SHORTSTRING("senduipi"),
    ZYDIS_MAKE_SHORTSTRING("serialize"),
    ZYDIS_MAKE_SHORTSTRING("setb"),
    ZYDIS_MAKE_SHORTSTRING("setbe"),
    ZYDIS_MAKE_SHORTSTRING("setl"),
    ZYDIS_MAKE_SHORTSTRING("setle"),
    ZYDIS_MAKE_SHORTSTRING("setnb"),
    ZYDIS_MAKE_SHORTSTRING("setnbe"),
    ZYDIS_MAKE_SHORTSTRING("setnl"),
    ZYDIS_MAKE_SHORTSTRING("setnle"),
    ZYDIS_MAKE_SHORTSTRING("setno"),
    ZYDIS_MAKE_SHORTSTRING("setnp"),
    ZYDIS_MAKE_SHORTSTRING("setns"),
    ZYDIS_MAKE_SHORTSTRING("setnz"),
    ZYDIS_MAKE_SHORTSTRING("seto"),
    ZYDIS_MAKE_SHORTSTRING("setp"),
    ZYDIS_MAKE_SHORTSTRING("sets"),
    ZYDIS_MAKE_SHORTSTRING("setssbsy"),
    ZYDIS_MAKE_SHORTSTRING("setz"),
    ZYDIS_MAKE_SHORTSTRING("sfence"),
    ZYDIS_MAKE_SHORTSTRING("sgdt"),
    ZYDIS_MAKE_SHORTSTRING("sha1msg1"),
    ZYDIS_MAKE_SHORTSTRING("sha1msg2"),
    ZYDIS_MAKE_SHORTSTRING("sha1nexte"),
    ZYDIS_MAKE_SHORTSTRING("sha1rnds4"),
    ZYDIS_MAKE_SHORTSTRING("sha256msg1"),
    ZYDIS_MAKE_SHORTSTRING("sha256msg2"),
    ZYDIS_MAKE_SHORTSTRING("sha256rnds2"),
    ZYDIS_MAKE_SHORTSTRING("shl"),
    ZYDIS_MAKE_SHORTSTRING("shld"),
    ZYDIS_MAKE_SHORTSTRING("shlx"),
    ZYDIS_MAKE_SHORTSTRING("shr"),
    ZYDIS_MAKE_SHORTSTRING("shrd"),
    ZYDIS_MAKE_SHORTSTRING("shrx"),
    ZYDIS_MAKE_SHORTSTRING("shufpd"),
    ZYDIS_MAKE_SHORTSTRING("shufps"),
    ZYDIS_MAKE_SHORTSTRING("sidt"),
    ZYDIS_MAKE_SHORTSTRING("skinit"),
    ZYDIS_MAKE_SHORTSTRING("sldt"),
    ZYDIS_MAKE_SHORTSTRING("slwpcb"),
    ZYDIS_MAKE_SHORTSTRING("smsw"),
    ZYDIS_MAKE_SHORTSTRING("spflt"),
    ZYDIS_MAKE_SHORTSTRING("sqrtpd"),
    ZYDIS_MAKE_SHORTSTRING("sqrtps"),
    ZYDIS_MAKE_SHORTSTRING("sqrtsd"),
    ZYDIS_MAKE_SHORTSTRING("sqrtss"),
    ZYDIS_MAKE_SHORTSTRING("stac"),
    ZYDIS_MAKE_SHORTSTRING("stc"),
    ZYDIS_MAKE_SHORTSTRING("std"),
    ZYDIS_MAKE_SHORTSTRING("stgi"),
    ZYDIS_MAKE_SHORTSTRING("sti"),
    ZYDIS_MAKE_SHORTSTRING("stmxcsr"),
    ZYDIS_MAKE_SHORTSTRING("stosb"),
    ZYDIS_MAKE_SHORTSTRING("stosd"),
    ZYDIS_MAKE_SHORTSTRING("stosq"),
    ZYDIS_MAKE_SHORTSTRING("stosw"),
    ZYDIS_MAKE_SHORTSTRING("str"),
    ZYDIS_MAKE_SHORTSTRING("sttilecfg"),
    ZYDIS_MAKE_SHORTSTRING("stui"),
    ZYDIS_MAKE_SHORTSTRING("sub"),
    ZYDIS_MAKE_SHORTSTRING("subpd"),
    ZYDIS_MAKE_SHORTSTRING("subps"),
    ZYDIS_MAKE_SHORTSTRING("subsd"),
    ZYDIS_MAKE_SHORTSTRING("subss"),
    ZYDIS_MAKE_SHORTSTRING("swapgs"),
    ZYDIS_MAKE_SHORTSTRING("syscall"),
    ZYDIS_MAKE_SHORTSTRING("sysenter"),
    ZYDIS_MAKE_SHORTSTRING("sysexit"),
    ZYDIS_MAKE_SHORTSTRING("sysret"),
    ZYDIS_MAKE_SHORTSTRING("t1mskc"),
    ZYDIS_MAKE_SHORTSTRING("tdcall"),
    ZYDIS_MAKE_SHORTSTRING("tdpbf16ps"),
    ZYDIS_MAKE_SHORTSTRING("tdpbssd"),
    ZYDIS_MAKE_SHORTSTRING("tdpbsud"),
    ZYDIS_MAKE_SHORTSTRING("tdpbusd"),
    ZYDIS_MAKE_SHORTSTRING("tdpbuud"),
    ZYDIS_MAKE_SHORTSTRING("test"),
    ZYDIS_MAKE_SHORTSTRING("testui"),
    ZYDIS_MAKE_SHORTSTRING("tileloadd"),
    ZYDIS_MAKE_SHORTSTRING("tileloaddt1"),
    ZYDIS_MAKE_SHORTSTRING("tilerelease"),
    ZYDIS_MAKE_SHORTSTRING("tilestored"),
    ZYDIS_MAKE_SHORTSTRING("tilezero"),
    ZYDIS_MAKE_SHORTSTRING("tlbsync"),
    ZYDIS_MAKE_SHORTSTRING("tpause"),
    ZYDIS_MAKE_SHORTSTRING("tzcnt"),
    ZYDIS_MAKE_SHORTSTRING("tzcnti"),
    ZYDIS_MAKE_SHORTSTRING("tzmsk"),
    ZYDIS_MAKE_SHORTSTRING("ucomisd"),
    ZYDIS_MAKE_SHORTSTRING("ucomiss"),
    ZYDIS_MAKE_SHORTSTRING("ud0"),
    ZYDIS_MAKE_SHORTSTRING("ud1"),
    ZYDIS_MAKE_SHORTSTRING("ud2"),
    ZYDIS_MAKE_SHORTSTRING("uiret"),
    ZYDIS_MAKE_SHORTSTRING("umonitor"),
    ZYDIS_MAKE_SHORTSTRING("umwait"),
    ZYDIS_MAKE_SHORTSTRING("unpckhpd"),
    ZYDIS_MAKE_SHORTSTRING("unpckhps"),
    ZYDIS_MAKE_SHORTSTRING("unpcklpd"),
    ZYDIS_MAKE_SHORTSTRING("unpcklps"),
    ZYDIS_MAKE_SHORTSTRING("v4fmaddps"),
    ZYDIS_MAKE_SHORTSTRING("v4fmaddss"),
    ZYDIS_MAKE_SHORTSTRING("v4fnmaddps"),
    ZYDIS_MAKE_SHORTSTRING("v4fnmaddss"),
    ZYDIS_MAKE_SHORTSTRING("vaddnpd"),
    ZYDIS_MAKE_SHORTSTRING("vaddnps"),
    ZYDIS_MAKE_SHORTSTRING("vaddpd"),
    ZYDIS_MAKE_SHORTSTRING("vaddph"),
    ZYDIS_MAKE_SHORTSTRING("vaddps"),
    ZYDIS_MAKE_SHORTSTRING("vaddsd"),
    ZYDIS_MAKE_SHORTSTRING("vaddsetsps"),
    ZYDIS_MAKE_SHORTSTRING("vaddsh"),
    ZYDIS_MAKE_SHORTSTRING("vaddss"),
    ZYDIS_MAKE_SHORTSTRING("vaddsubpd"),
    ZYDIS_MAKE_SHORTSTRING("vaddsubps"),
    ZYDIS_MAKE_SHORTSTRING("vaesdec"),
    ZYDIS_MAKE_SHORTSTRING("vaesdeclast"),
    ZYDIS_MAKE_SHORTSTRING("vaesenc"),
    ZYDIS_MAKE_SHORTSTRING("vaesenclast"),
    ZYDIS_MAKE_SHORTSTRING("vaesimc"),
    ZYDIS_MAKE_SHORTSTRING("vaeskeygenassist"),
    ZYDIS_MAKE_SHORTSTRING("valignd"),
    ZYDIS_MAKE_SHORTSTRING("valignq"),
    ZYDIS_MAKE_SHORTSTRING("vandnpd"),
    ZYDIS_MAKE_SHORTSTRING("vandnps"),
    ZYDIS_MAKE_SHORTSTRING("vandpd"),
    ZYDIS_MAKE_SHORTSTRING("vandps"),
    ZYDIS_MAKE_SHORTSTRING("vblendmpd"),
    ZYDIS_MAKE_SHORTSTRING("vblendmps"),
    ZYDIS_MAKE_SHORTSTRING("vblendpd"),
    ZYDIS_MAKE_SHORTSTRING("vblendps"),
    ZYDIS_MAKE_SHORTSTRING("vblendvpd"),
    ZYDIS_MAKE_SHORTSTRING("vblendvps"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastf128"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastf32x2"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastf32x4"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastf32x8"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastf64x2"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastf64x4"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcasti128"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcasti32x2"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcasti32x4"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcasti32x8"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcasti64x2"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcasti64x4"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastsd"),
    ZYDIS_MAKE_SHORTSTRING("vbroadcastss"),
    ZYDIS_MAKE_SHORTSTRING("vcmppd"),
    ZYDIS_MAKE_SHORTSTRING("vcmpph"),
    ZYDIS_MAKE_SHORTSTRING("vcmpps"),
    ZYDIS_MAKE_SHORTSTRING("vcmpsd"),
    ZYDIS_MAKE_SHORTSTRING("vcmpsh"),
    ZYDIS_MAKE_SHORTSTRING("vcmpss"),
    ZYDIS_MAKE_SHORTSTRING("vcomisd"),
    ZYDIS_MAKE_SHORTSTRING("vcomish"),
    ZYDIS_MAKE_SHORTSTRING("vcomiss"),
    ZYDIS_MAKE_SHORTSTRING("vcompresspd"),
    ZYDIS_MAKE_SHORTSTRING("vcompressps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtdq2pd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtdq2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtdq2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtfxpntdq2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtfxpntpd2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtfxpntpd2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtfxpntps2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtfxpntps2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtfxpntudq2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtne2ps2bf16"),
    ZYDIS_MAKE_SHORTSTRING("vcvtneps2bf16"),
    ZYDIS_MAKE_SHORTSTRING("vcvtpd2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtpd2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtpd2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtpd2qq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtpd2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtpd2uqq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2pd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2psx"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2qq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2uqq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2uw"),
    ZYDIS_MAKE_SHORTSTRING("vcvtph2w"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2pd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2phx"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2qq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtps2uqq"),
    ZYDIS_MAKE_SHORTSTRING("vcvtqq2pd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtqq2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtqq2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsd2sh"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsd2si"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsd2ss"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsd2usi"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsh2sd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsh2si"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsh2ss"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsh2usi"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsi2sd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsi2sh"),
    ZYDIS_MAKE_SHORTSTRING("vcvtsi2ss"),
    ZYDIS_MAKE_SHORTSTRING("vcvtss2sd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtss2sh"),
    ZYDIS_MAKE_SHORTSTRING("vcvtss2si"),
    ZYDIS_MAKE_SHORTSTRING("vcvtss2usi"),
    ZYDIS_MAKE_SHORTSTRING("vcvttpd2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttpd2qq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttpd2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttpd2uqq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttph2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttph2qq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttph2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttph2uqq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttph2uw"),
    ZYDIS_MAKE_SHORTSTRING("vcvttph2w"),
    ZYDIS_MAKE_SHORTSTRING("vcvttps2dq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttps2qq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttps2udq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttps2uqq"),
    ZYDIS_MAKE_SHORTSTRING("vcvttsd2si"),
    ZYDIS_MAKE_SHORTSTRING("vcvttsd2usi"),
    ZYDIS_MAKE_SHORTSTRING("vcvttsh2si"),
    ZYDIS_MAKE_SHORTSTRING("vcvttsh2usi"),
    ZYDIS_MAKE_SHORTSTRING("vcvttss2si"),
    ZYDIS_MAKE_SHORTSTRING("vcvttss2usi"),
    ZYDIS_MAKE_SHORTSTRING("vcvtudq2pd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtudq2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtudq2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtuqq2pd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtuqq2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtuqq2ps"),
    ZYDIS_MAKE_SHORTSTRING("vcvtusi2sd"),
    ZYDIS_MAKE_SHORTSTRING("vcvtusi2sh"),
    ZYDIS_MAKE_SHORTSTRING("vcvtusi2ss"),
    ZYDIS_MAKE_SHORTSTRING("vcvtuw2ph"),
    ZYDIS_MAKE_SHORTSTRING("vcvtw2ph"),
    ZYDIS_MAKE_SHORTSTRING("vdbpsadbw"),
    ZYDIS_MAKE_SHORTSTRING("vdivpd"),
    ZYDIS_MAKE_SHORTSTRING("vdivph"),
    ZYDIS_MAKE_SHORTSTRING("vdivps"),
    ZYDIS_MAKE_SHORTSTRING("vdivsd"),
    ZYDIS_MAKE_SHORTSTRING("vdivsh"),
    ZYDIS_MAKE_SHORTSTRING("vdivss"),
    ZYDIS_MAKE_SHORTSTRING("vdpbf16ps"),
    ZYDIS_MAKE_SHORTSTRING("vdppd"),
    ZYDIS_MAKE_SHORTSTRING("vdpps"),
    ZYDIS_MAKE_SHORTSTRING("verr"),
    ZYDIS_MAKE_SHORTSTRING("verw"),
    ZYDIS_MAKE_SHORTSTRING("vexp223ps"),
    ZYDIS_MAKE_SHORTSTRING("vexp2pd"),
    ZYDIS_MAKE_SHORTSTRING("vexp2ps"),
    ZYDIS_MAKE_SHORTSTRING("vexpandpd"),
    ZYDIS_MAKE_SHORTSTRING("vexpandps"),
    ZYDIS_MAKE_SHORTSTRING("vextractf128"),
    ZYDIS_MAKE_SHORTSTRING("vextractf32x4"),
    ZYDIS_MAKE_SHORTSTRING("vextractf32x8"),
    ZYDIS_MAKE_SHORTSTRING("vextractf64x2"),
    ZYDIS_MAKE_SHORTSTRING("vextractf64x4"),
    ZYDIS_MAKE_SHORTSTRING("vextracti128"),
    ZYDIS_MAKE_SHORTSTRING("vextracti32x4"),
    ZYDIS_MAKE_SHORTSTRING("vextracti32x8"),
    ZYDIS_MAKE_SHORTSTRING("vextracti64x2"),
    ZYDIS_MAKE_SHORTSTRING("vextracti64x4"),
    ZYDIS_MAKE_SHORTSTRING("vextractps"),
    ZYDIS_MAKE_SHORTSTRING("vfcmaddcph"),
    ZYDIS_MAKE_SHORTSTRING("vfcmaddcsh"),
    ZYDIS_MAKE_SHORTSTRING("vfcmulcph"),
    ZYDIS_MAKE_SHORTSTRING("vfcmulcsh"),
    ZYDIS_MAKE_SHORTSTRING("vfixupimmpd"),
    ZYDIS_MAKE_SHORTSTRING("vfixupimmps"),
    ZYDIS_MAKE_SHORTSTRING("vfixupimmsd"),
    ZYDIS_MAKE_SHORTSTRING("vfixupimmss"),
    ZYDIS_MAKE_SHORTSTRING("vfixupnanpd"),
    ZYDIS_MAKE_SHORTSTRING("vfixupnanps"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd132pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd132ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd132ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd132sd"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd132sh"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd132ss"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd213pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd213ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd213ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd213sd"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd213sh"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd213ss"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd231pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd231ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd231ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd231sd"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd231sh"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd231ss"),
    ZYDIS_MAKE_SHORTSTRING("vfmadd233ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddcph"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddcsh"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddpd"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddps"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsd"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddss"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub132pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub132ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub132ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub213pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub213ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub213ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub231pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub231ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsub231ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsubpd"),
    ZYDIS_MAKE_SHORTSTRING("vfmaddsubps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub132pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub132ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub132ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub132sd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub132sh"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub132ss"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub213pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub213ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub213ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub213sd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub213sh"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub213ss"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub231pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub231ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub231ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub231sd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub231sh"),
    ZYDIS_MAKE_SHORTSTRING("vfmsub231ss"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd132pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd132ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd132ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd213pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd213ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd213ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd231pd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd231ph"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubadd231ps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubaddpd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubaddps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubpd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubps"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubsd"),
    ZYDIS_MAKE_SHORTSTRING("vfmsubss"),
    ZYDIS_MAKE_SHORTSTRING("vfmulcph"),
    ZYDIS_MAKE_SHORTSTRING("vfmulcsh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd132pd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd132ph"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd132ps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd132sd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd132sh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd132ss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd213pd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd213ph"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd213ps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd213sd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd213sh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd213ss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd231pd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd231ph"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd231ps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd231sd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd231sh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmadd231ss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmaddpd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmaddps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmaddsd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmaddss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub132pd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub132ph"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub132ps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub132sd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub132sh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub132ss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub213pd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub213ph"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub213ps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub213sd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub213sh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub213ss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub231pd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub231ph"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub231ps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub231sd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub231sh"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsub231ss"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsubpd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsubps"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsubsd"),
    ZYDIS_MAKE_SHORTSTRING("vfnmsubss"),
    ZYDIS_MAKE_SHORTSTRING("vfpclasspd"),
    ZYDIS_MAKE_SHORTSTRING("vfpclassph"),
    ZYDIS_MAKE_SHORTSTRING("vfpclassps"),
    ZYDIS_MAKE_SHORTSTRING("vfpclasssd"),
    ZYDIS_MAKE_SHORTSTRING("vfpclasssh"),
    ZYDIS_MAKE_SHORTSTRING("vfpclassss"),
    ZYDIS_MAKE_SHORTSTRING("vfrczpd"),
    ZYDIS_MAKE_SHORTSTRING("vfrczps"),
    ZYDIS_MAKE_SHORTSTRING("vfrczsd"),
    ZYDIS_MAKE_SHORTSTRING("vfrczss"),
    ZYDIS_MAKE_SHORTSTRING("vgatherdpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherdps"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf0dpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf0dps"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf0hintdpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf0hintdps"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf0qpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf0qps"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf1dpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf1dps"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf1qpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherpf1qps"),
    ZYDIS_MAKE_SHORTSTRING("vgatherqpd"),
    ZYDIS_MAKE_SHORTSTRING("vgatherqps"),
    ZYDIS_MAKE_SHORTSTRING("vgetexppd"),
    ZYDIS_MAKE_SHORTSTRING("vgetexpph"),
    ZYDIS_MAKE_SHORTSTRING("vgetexpps"),
    ZYDIS_MAKE_SHORTSTRING("vgetexpsd"),
    ZYDIS_MAKE_SHORTSTRING("vgetexpsh"),
    ZYDIS_MAKE_SHORTSTRING("vgetexpss"),
    ZYDIS_MAKE_SHORTSTRING("vgetmantpd"),
    ZYDIS_MAKE_SHORTSTRING("vgetmantph"),
    ZYDIS_MAKE_SHORTSTRING("vgetmantps"),
    ZYDIS_MAKE_SHORTSTRING("vgetmantsd"),
    ZYDIS_MAKE_SHORTSTRING("vgetmantsh"),
    ZYDIS_MAKE_SHORTSTRING("vgetmantss"),
    ZYDIS_MAKE_SHORTSTRING("vgf2p8affineinvqb"),
    ZYDIS_MAKE_SHORTSTRING("vgf2p8affineqb"),
    ZYDIS_MAKE_SHORTSTRING("vgf2p8mulb"),
    ZYDIS_MAKE_SHORTSTRING("vgmaxabsps"),
    ZYDIS_MAKE_SHORTSTRING("vgmaxpd"),
    ZYDIS_MAKE_SHORTSTRING("vgmaxps"),
    ZYDIS_MAKE_SHORTSTRING("vgminpd"),
    ZYDIS_MAKE_SHORTSTRING("vgminps"),
    ZYDIS_MAKE_SHORTSTRING("vhaddpd"),
    ZYDIS_MAKE_SHORTSTRING("vhaddps"),
    ZYDIS_MAKE_SHORTSTRING("vhsubpd"),
    ZYDIS_MAKE_SHORTSTRING("vhsubps"),
    ZYDIS_MAKE_SHORTSTRING("vinsertf128"),
    ZYDIS_MAKE_SHORTSTRING("vinsertf32x4"),
    ZYDIS_MAKE_SHORTSTRING("vinsertf32x8"),
    ZYDIS_MAKE_SHORTSTRING("vinsertf64x2"),
    ZYDIS_MAKE_SHORTSTRING("vinsertf64x4"),
    ZYDIS_MAKE_SHORTSTRING("vinserti128"),
    ZYDIS_MAKE_SHORTSTRING("vinserti32x4"),
    ZYDIS_MAKE_SHORTSTRING("vinserti32x8"),
    ZYDIS_MAKE_SHORTSTRING("vinserti64x2"),
    ZYDIS_MAKE_SHORTSTRING("vinserti64x4"),
    ZYDIS_MAKE_SHORTSTRING("vinsertps"),
    ZYDIS_MAKE_SHORTSTRING("vlddqu"),
    ZYDIS_MAKE_SHORTSTRING("vldmxcsr"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpackhd"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpackhpd"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpackhps"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpackhq"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpackld"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpacklpd"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpacklps"),
    ZYDIS_MAKE_SHORTSTRING("vloadunpacklq"),
    ZYDIS_MAKE_SHORTSTRING("vlog2ps"),
    ZYDIS_MAKE_SHORTSTRING("vmaskmovdqu"),
    ZYDIS_MAKE_SHORTSTRING("vmaskmovpd"),
    ZYDIS_MAKE_SHORTSTRING("vmaskmovps"),
    ZYDIS_MAKE_SHORTSTRING("vmaxpd"),
    ZYDIS_MAKE_SHORTSTRING("vmaxph"),
    ZYDIS_MAKE_SHORTSTRING("vmaxps"),
    ZYDIS_MAKE_SHORTSTRING("vmaxsd"),
    ZYDIS_MAKE_SHORTSTRING("vmaxsh"),
    ZYDIS_MAKE_SHORTSTRING("vmaxss"),
    ZYDIS_MAKE_SHORTSTRING("vmcall"),
    ZYDIS_MAKE_SHORTSTRING("vmclear"),
    ZYDIS_MAKE_SHORTSTRING("vmfunc"),
    ZYDIS_MAKE_SHORTSTRING("vminpd"),
    ZYDIS_MAKE_SHORTSTRING("vminph"),
    ZYDIS_MAKE_SHORTSTRING("vminps"),
    ZYDIS_MAKE_SHORTSTRING("vminsd"),
    ZYDIS_MAKE_SHORTSTRING("vminsh"),
    ZYDIS_MAKE_SHORTSTRING("vminss"),
    ZYDIS_MAKE_SHORTSTRING("vmlaunch"),
    ZYDIS_MAKE_SHORTSTRING("vmload"),
    ZYDIS_MAKE_SHORTSTRING("vmmcall"),
    ZYDIS_MAKE_SHORTSTRING("vmovapd"),
    ZYDIS_MAKE_SHORTSTRING("vmovaps"),
    ZYDIS_MAKE_SHORTSTRING("vmovd"),
    ZYDIS_MAKE_SHORTSTRING("vmovddup"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqa"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqa32"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqa64"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqu"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqu16"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqu32"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqu64"),
    ZYDIS_MAKE_SHORTSTRING("vmovdqu8"),
    ZYDIS_MAKE_SHORTSTRING("vmovhlps"),
    ZYDIS_MAKE_SHORTSTRING("vmovhpd"),
    ZYDIS_MAKE_SHORTSTRING("vmovhps"),
    ZYDIS_MAKE_SHORTSTRING("vmovlhps"),
    ZYDIS_MAKE_SHORTSTRING("vmovlpd"),
    ZYDIS_MAKE_SHORTSTRING("vmovlps"),
    ZYDIS_MAKE_SHORTSTRING("vmovmskpd"),
    ZYDIS_MAKE_SHORTSTRING("vmovmskps"),
    ZYDIS_MAKE_SHORTSTRING("vmovnrapd"),
    ZYDIS_MAKE_SHORTSTRING("vmovnraps"),
    ZYDIS_MAKE_SHORTSTRING("vmovnrngoapd"),
    ZYDIS_MAKE_SHORTSTRING("vmovnrngoaps"),
    ZYDIS_MAKE_SHORTSTRING("vmovntdq"),
    ZYDIS_MAKE_SHORTSTRING("vmovntdqa"),
    ZYDIS_MAKE_SHORTSTRING("vmovntpd"),
    ZYDIS_MAKE_SHORTSTRING("vmovntps"),
    ZYDIS_MAKE_SHORTSTRING("vmovq"),
    ZYDIS_MAKE_SHORTSTRING("vmovsd"),
    ZYDIS_MAKE_SHORTSTRING("vmovsh"),
    ZYDIS_MAKE_SHORTSTRING("vmovshdup"),
    ZYDIS_MAKE_SHORTSTRING("vmovsldup"),
    ZYDIS_MAKE_SHORTSTRING("vmovss"),
    ZYDIS_MAKE_SHORTSTRING("vmovupd"),
    ZYDIS_MAKE_SHORTSTRING("vmovups"),
    ZYDIS_MAKE_SHORTSTRING("vmovw"),
    ZYDIS_MAKE_SHORTSTRING("vmpsadbw"),
    ZYDIS_MAKE_SHORTSTRING("vmptrld"),
    ZYDIS_MAKE_SHORTSTRING("vmptrst"),
    ZYDIS_MAKE_SHORTSTRING("vmread"),
    ZYDIS_MAKE_SHORTSTRING("vmresume"),
    ZYDIS_MAKE_SHORTSTRING("vmrun"),
    ZYDIS_MAKE_SHORTSTRING("vmsave"),
    ZYDIS_MAKE_SHORTSTRING("vmulpd"),
    ZYDIS_MAKE_SHORTSTRING("vmulph"),
    ZYDIS_MAKE_SHORTSTRING("vmulps"),
    ZYDIS_MAKE_SHORTSTRING("vmulsd"),
    ZYDIS_MAKE_SHORTSTRING("vmulsh"),
    ZYDIS_MAKE_SHORTSTRING("vmulss"),
    ZYDIS_MAKE_SHORTSTRING("vmwrite"),
    ZYDIS_MAKE_SHORTSTRING("vmxoff"),
    ZYDIS_MAKE_SHORTSTRING("vmxon"),
    ZYDIS_MAKE_SHORTSTRING("vorpd"),
    ZYDIS_MAKE_SHORTSTRING("vorps"),
    ZYDIS_MAKE_SHORTSTRING("vp2intersectd"),
    ZYDIS_MAKE_SHORTSTRING("vp2intersectq"),
    ZYDIS_MAKE_SHORTSTRING("vp4dpwssd"),
    ZYDIS_MAKE_SHORTSTRING("vp4dpwssds"),
    ZYDIS_MAKE_SHORTSTRING("vpabsb"),
    ZYDIS_MAKE_SHORTSTRING("vpabsd"),
    ZYDIS_MAKE_SHORTSTRING("vpabsq"),
    ZYDIS_MAKE_SHORTSTRING("vpabsw"),
    ZYDIS_MAKE_SHORTSTRING("vpackssdw"),
    ZYDIS_MAKE_SHORTSTRING("vpacksswb"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorehd"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorehpd"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorehps"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorehq"),
    ZYDIS_MAKE_SHORTSTRING("vpackstoreld"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorelpd"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorelps"),
    ZYDIS_MAKE_SHORTSTRING("vpackstorelq"),
    ZYDIS_MAKE_SHORTSTRING("vpackusdw"),
    ZYDIS_MAKE_SHORTSTRING("vpackuswb"),
    ZYDIS_MAKE_SHORTSTRING("vpadcd"),
    ZYDIS_MAKE_SHORTSTRING("vpaddb"),
    ZYDIS_MAKE_SHORTSTRING("vpaddd"),
    ZYDIS_MAKE_SHORTSTRING("vpaddq"),
    ZYDIS_MAKE_SHORTSTRING("vpaddsb"),
    ZYDIS_MAKE_SHORTSTRING("vpaddsetcd"),
    ZYDIS_MAKE_SHORTSTRING("vpaddsetsd"),
    ZYDIS_MAKE_SHORTSTRING("vpaddsw"),
    ZYDIS_MAKE_SHORTSTRING("vpaddusb"),
    ZYDIS_MAKE_SHORTSTRING("vpaddusw"),
    ZYDIS_MAKE_SHORTSTRING("vpaddw"),
    ZYDIS_MAKE_SHORTSTRING("vpalignr"),
    ZYDIS_MAKE_SHORTSTRING("vpand"),
    ZYDIS_MAKE_SHORTSTRING("vpandd"),
    ZYDIS_MAKE_SHORTSTRING("vpandn"),
    ZYDIS_MAKE_SHORTSTRING("vpandnd"),
    ZYDIS_MAKE_SHORTSTRING("vpandnq"),
    ZYDIS_MAKE_SHORTSTRING("vpandq"),
    ZYDIS_MAKE_SHORTSTRING("vpavgb"),
    ZYDIS_MAKE_SHORTSTRING("vpavgw"),
    ZYDIS_MAKE_SHORTSTRING("vpblendd"),
    ZYDIS_MAKE_SHORTSTRING("vpblendmb"),
    ZYDIS_MAKE_SHORTSTRING("vpblendmd"),
    ZYDIS_MAKE_SHORTSTRING("vpblendmq"),
    ZYDIS_MAKE_SHORTSTRING("vpblendmw"),
    ZYDIS_MAKE_SHORTSTRING("vpblendvb"),
    ZYDIS_MAKE_SHORTSTRING("vpblendw"),
    ZYDIS_MAKE_SHORTSTRING("vpbroadcastb"),
    ZYDIS_MAKE_SHORTSTRING("vpbroadcastd"),
    ZYDIS_MAKE_SHORTSTRING("vpbroadcastmb2q"),
    ZYDIS_MAKE_SHORTSTRING("vpbroadcastmw2d"),
    ZYDIS_MAKE_SHORTSTRING("vpbroadcastq"),
    ZYDIS_MAKE_SHORTSTRING("vpbroadcastw"),
    ZYDIS_MAKE_SHORTSTRING("vpclmulqdq"),
    ZYDIS_MAKE_SHORTSTRING("vpcmov"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpb"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpd"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpeqb"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpeqd"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpeqq"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpeqw"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpestri"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpestrm"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpgtb"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpgtd"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpgtq"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpgtw"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpistri"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpistrm"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpltd"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpq"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpub"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpud"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpuq"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpuw"),
    ZYDIS_MAKE_SHORTSTRING("vpcmpw"),
    ZYDIS_MAKE_SHORTSTRING("vpcomb"),
    ZYDIS_MAKE_SHORTSTRING("vpcomd"),
    ZYDIS_MAKE_SHORTSTRING("vpcompressb"),
    ZYDIS_MAKE_SHORTSTRING("vpcompressd"),
    ZYDIS_MAKE_SHORTSTRING("vpcompressq"),
    ZYDIS_MAKE_SHORTSTRING("vpcompressw"),
    ZYDIS_MAKE_SHORTSTRING("vpcomq"),
    ZYDIS_MAKE_SHORTSTRING("vpcomub"),
    ZYDIS_MAKE_SHORTSTRING("vpcomud"),
    ZYDIS_MAKE_SHORTSTRING("vpcomuq"),
    ZYDIS_MAKE_SHORTSTRING("vpcomuw"),
    ZYDIS_MAKE_SHORTSTRING("vpcomw"),
    ZYDIS_MAKE_SHORTSTRING("vpconflictd"),
    ZYDIS_MAKE_SHORTSTRING("vpconflictq"),
    ZYDIS_MAKE_SHORTSTRING("vpdpbusd"),
    ZYDIS_MAKE_SHORTSTRING("vpdpbusds"),
    ZYDIS_MAKE_SHORTSTRING("vpdpwssd"),
    ZYDIS_MAKE_SHORTSTRING("vpdpwssds"),
    ZYDIS_MAKE_SHORTSTRING("vperm2f128"),
    ZYDIS_MAKE_SHORTSTRING("vperm2i128"),
    ZYDIS_MAKE_SHORTSTRING("vpermb"),
    ZYDIS_MAKE_SHORTSTRING("vpermd"),
    ZYDIS_MAKE_SHORTSTRING("vpermf32x4"),
    ZYDIS_MAKE_SHORTSTRING("vpermi2b"),
    ZYDIS_MAKE_SHORTSTRING("vpermi2d"),
    ZYDIS_MAKE_SHORTSTRING("vpermi2pd"),
    ZYDIS_MAKE_SHORTSTRING("vpermi2ps"),
    ZYDIS_MAKE_SHORTSTRING("vpermi2q"),
    ZYDIS_MAKE_SHORTSTRING("vpermi2w"),
    ZYDIS_MAKE_SHORTSTRING("vpermil2pd"),
    ZYDIS_MAKE_SHORTSTRING("vpermil2ps"),
    ZYDIS_MAKE_SHORTSTRING("vpermilpd"),
    ZYDIS_MAKE_SHORTSTRING("vpermilps"),
    ZYDIS_MAKE_SHORTSTRING("vpermpd"),
    ZYDIS_MAKE_SHORTSTRING("vpermps"),
    ZYDIS_MAKE_SHORTSTRING("vpermq"),
    ZYDIS_MAKE_SHORTSTRING("vpermt2b"),
    ZYDIS_MAKE_SHORTSTRING("vpermt2d"),
    ZYDIS_MAKE_SHORTSTRING("vpermt2pd"),
    ZYDIS_MAKE_SHORTSTRING("vpermt2ps"),
    ZYDIS_MAKE_SHORTSTRING("vpermt2q"),
    ZYDIS_MAKE_SHORTSTRING("vpermt2w"),
    ZYDIS_MAKE_SHORTSTRING("vpermw"),
    ZYDIS_MAKE_SHORTSTRING("vpexpandb"),
    ZYDIS_MAKE_SHORTSTRING("vpexpandd"),
    ZYDIS_MAKE_SHORTSTRING("vpexpandq"),
    ZYDIS_MAKE_SHORTSTRING("vpexpandw"),
    ZYDIS_MAKE_SHORTSTRING("vpextrb"),
    ZYDIS_MAKE_SHORTSTRING("vpextrd"),
    ZYDIS_MAKE_SHORTSTRING("vpextrq"),
    ZYDIS_MAKE_SHORTSTRING("vpextrw"),
    ZYDIS_MAKE_SHORTSTRING("vpgatherdd"),
    ZYDIS_MAKE_SHORTSTRING("vpgatherdq"),
    ZYDIS_MAKE_SHORTSTRING("vpgatherqd"),
    ZYDIS_MAKE_SHORTSTRING("vpgatherqq"),
    ZYDIS_MAKE_SHORTSTRING("vphaddbd"),
    ZYDIS_MAKE_SHORTSTRING("vphaddbq"),
    ZYDIS_MAKE_SHORTSTRING("vphaddbw"),
    ZYDIS_MAKE_SHORTSTRING("vphaddd"),
    ZYDIS_MAKE_SHORTSTRING("vphadddq"),
    ZYDIS_MAKE_SHORTSTRING("vphaddsw"),
    ZYDIS_MAKE_SHORTSTRING("vphaddubd"),
    ZYDIS_MAKE_SHORTSTRING("vphaddubq"),
    ZYDIS_MAKE_SHORTSTRING("vphaddubw"),
    ZYDIS_MAKE_SHORTSTRING("vphaddudq"),
    ZYDIS_MAKE_SHORTSTRING("vphadduwd"),
    ZYDIS_MAKE_SHORTSTRING("vphadduwq"),
    ZYDIS_MAKE_SHORTSTRING("vphaddw"),
    ZYDIS_MAKE_SHORTSTRING("vphaddwd"),
    ZYDIS_MAKE_SHORTSTRING("vphaddwq"),
    ZYDIS_MAKE_SHORTSTRING("vphminposuw"),
    ZYDIS_MAKE_SHORTSTRING("vphsubbw"),
    ZYDIS_MAKE_SHORTSTRING("vphsubd"),
    ZYDIS_MAKE_SHORTSTRING("vphsubdq"),
    ZYDIS_MAKE_SHORTSTRING("vphsubsw"),
    ZYDIS_MAKE_SHORTSTRING("vphsubw"),
    ZYDIS_MAKE_SHORTSTRING("vphsubwd"),
    ZYDIS_MAKE_SHORTSTRING("vpinsrb"),
    ZYDIS_MAKE_SHORTSTRING("vpinsrd"),
    ZYDIS_MAKE_SHORTSTRING("vpinsrq"),
    ZYDIS_MAKE_SHORTSTRING("vpinsrw"),
    ZYDIS_MAKE_SHORTSTRING("vplzcntd"),
    ZYDIS_MAKE_SHORTSTRING("vplzcntq"),
    ZYDIS_MAKE_SHORTSTRING("vpmacsdd"),
    ZYDIS_MAKE_SHORTSTRING("vpmacsdqh"),
    ZYDIS_MAKE_SHORTSTRING("vpmacsdql"),
    ZYDIS_MAKE_SHORTSTRING("vpmacssdd"),
    ZYDIS_MAKE_SHORTSTRING("vpmacssdqh"),
    ZYDIS_MAKE_SHORTSTRING("vpmacssdql"),
    ZYDIS_MAKE_SHORTSTRING("vpmacsswd"),
    ZYDIS_MAKE_SHORTSTRING("vpmacssww"),
    ZYDIS_MAKE_SHORTSTRING("vpmacswd"),
    ZYDIS_MAKE_SHORTSTRING("vpmacsww"),
    ZYDIS_MAKE_SHORTSTRING("vpmadcsswd"),
    ZYDIS_MAKE_SHORTSTRING("vpmadcswd"),
    ZYDIS_MAKE_SHORTSTRING("vpmadd231d"),
    ZYDIS_MAKE_SHORTSTRING("vpmadd233d"),
    ZYDIS_MAKE_SHORTSTRING("vpmadd52huq"),
    ZYDIS_MAKE_SHORTSTRING("vpmadd52luq"),
    ZYDIS_MAKE_SHORTSTRING("vpmaddubsw"),
    ZYDIS_MAKE_SHORTSTRING("vpmaddwd"),
    ZYDIS_MAKE_SHORTSTRING("vpmaskmovd"),
    ZYDIS_MAKE_SHORTSTRING("vpmaskmovq"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxsb"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxsd"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxsq"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxsw"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxub"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxud"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxuq"),
    ZYDIS_MAKE_SHORTSTRING("vpmaxuw"),
    ZYDIS_MAKE_SHORTSTRING("vpminsb"),
    ZYDIS_MAKE_SHORTSTRING("vpminsd"),
    ZYDIS_MAKE_SHORTSTRING("vpminsq"),
    ZYDIS_MAKE_SHORTSTRING("vpminsw"),
    ZYDIS_MAKE_SHORTSTRING("vpminub"),
    ZYDIS_MAKE_SHORTSTRING("vpminud"),
    ZYDIS_MAKE_SHORTSTRING("vpminuq"),
    ZYDIS_MAKE_SHORTSTRING("vpminuw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovb2m"),
    ZYDIS_MAKE_SHORTSTRING("vpmovd2m"),
    ZYDIS_MAKE_SHORTSTRING("vpmovdb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovdw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovm2b"),
    ZYDIS_MAKE_SHORTSTRING("vpmovm2d"),
    ZYDIS_MAKE_SHORTSTRING("vpmovm2q"),
    ZYDIS_MAKE_SHORTSTRING("vpmovm2w"),
    ZYDIS_MAKE_SHORTSTRING("vpmovmskb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovq2m"),
    ZYDIS_MAKE_SHORTSTRING("vpmovqb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovqd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovqw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsdb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsdw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsqb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsqd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsqw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovswb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsxbd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsxbq"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsxbw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsxdq"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsxwd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovsxwq"),
    ZYDIS_MAKE_SHORTSTRING("vpmovusdb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovusdw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovusqb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovusqd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovusqw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovuswb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovw2m"),
    ZYDIS_MAKE_SHORTSTRING("vpmovwb"),
    ZYDIS_MAKE_SHORTSTRING("vpmovzxbd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovzxbq"),
    ZYDIS_MAKE_SHORTSTRING("vpmovzxbw"),
    ZYDIS_MAKE_SHORTSTRING("vpmovzxdq"),
    ZYDIS_MAKE_SHORTSTRING("vpmovzxwd"),
    ZYDIS_MAKE_SHORTSTRING("vpmovzxwq"),
    ZYDIS_MAKE_SHORTSTRING("vpmuldq"),
    ZYDIS_MAKE_SHORTSTRING("vpmulhd"),
    ZYDIS_MAKE_SHORTSTRING("vpmulhrsw"),
    ZYDIS_MAKE_SHORTSTRING("vpmulhud"),
    ZYDIS_MAKE_SHORTSTRING("vpmulhuw"),
    ZYDIS_MAKE_SHORTSTRING("vpmulhw"),
    ZYDIS_MAKE_SHORTSTRING("vpmulld"),
    ZYDIS_MAKE_SHORTSTRING("vpmullq"),
    ZYDIS_MAKE_SHORTSTRING("vpmullw"),
    ZYDIS_MAKE_SHORTSTRING("vpmultishiftqb"),
    ZYDIS_MAKE_SHORTSTRING("vpmuludq"),
    ZYDIS_MAKE_SHORTSTRING("vpopcntb"),
    ZYDIS_MAKE_SHORTSTRING("vpopcntd"),
    ZYDIS_MAKE_SHORTSTRING("vpopcntq"),
    ZYDIS_MAKE_SHORTSTRING("vpopcntw"),
    ZYDIS_MAKE_SHORTSTRING("vpor"),
    ZYDIS_MAKE_SHORTSTRING("vpord"),
    ZYDIS_MAKE_SHORTSTRING("vporq"),
    ZYDIS_MAKE_SHORTSTRING("vpperm"),
    ZYDIS_MAKE_SHORTSTRING("vprefetch0"),
    ZYDIS_MAKE_SHORTSTRING("vprefetch1"),
    ZYDIS_MAKE_SHORTSTRING("vprefetch2"),
    ZYDIS_MAKE_SHORTSTRING("vprefetche0"),
    ZYDIS_MAKE_SHORTSTRING("vprefetche1"),
    ZYDIS_MAKE_SHORTSTRING("vprefetche2"),
    ZYDIS_MAKE_SHORTSTRING("vprefetchenta"),
    ZYDIS_MAKE_SHORTSTRING("vprefetchnta"),
    ZYDIS_MAKE_SHORTSTRING("vprold"),
    ZYDIS_MAKE_SHORTSTRING("vprolq"),
    ZYDIS_MAKE_SHORTSTRING("vprolvd"),
    ZYDIS_MAKE_SHORTSTRING("vprolvq"),
    ZYDIS_MAKE_SHORTSTRING("vprord"),
    ZYDIS_MAKE_SHORTSTRING("vprorq"),
    ZYDIS_MAKE_SHORTSTRING("vprorvd"),
    ZYDIS_MAKE_SHORTSTRING("vprorvq"),
    ZYDIS_MAKE_SHORTSTRING("vprotb"),
    ZYDIS_MAKE_SHORTSTRING("vprotd"),
    ZYDIS_MAKE_SHORTSTRING("vprotq"),
    ZYDIS_MAKE_SHORTSTRING("vprotw"),
    ZYDIS_MAKE_SHORTSTRING("vpsadbw"),
    ZYDIS_MAKE_SHORTSTRING("vpsbbd"),
    ZYDIS_MAKE_SHORTSTRING("vpsbbrd"),
    ZYDIS_MAKE_SHORTSTRING("vpscatterdd"),
    ZYDIS_MAKE_SHORTSTRING("vpscatterdq"),
    ZYDIS_MAKE_SHORTSTRING("vpscatterqd"),
    ZYDIS_MAKE_SHORTSTRING("vpscatterqq"),
    ZYDIS_MAKE_SHORTSTRING("vpshab"),
    ZYDIS_MAKE_SHORTSTRING("vpshad"),
    ZYDIS_MAKE_SHORTSTRING("vpshaq"),
    ZYDIS_MAKE_SHORTSTRING("vpshaw"),
    ZYDIS_MAKE_SHORTSTRING("vpshlb"),
    ZYDIS_MAKE_SHORTSTRING("vpshld"),
    ZYDIS_MAKE_SHORTSTRING("vpshldd"),
    ZYDIS_MAKE_SHORTSTRING("vpshldq"),
    ZYDIS_MAKE_SHORTSTRING("vpshldvd"),
    ZYDIS_MAKE_SHORTSTRING("vpshldvq"),
    ZYDIS_MAKE_SHORTSTRING("vpshldvw"),
    ZYDIS_MAKE_SHORTSTRING("vpshldw"),
    ZYDIS_MAKE_SHORTSTRING("vpshlq"),
    ZYDIS_MAKE_SHORTSTRING("vpshlw"),
    ZYDIS_MAKE_SHORTSTRING("vpshrdd"),
    ZYDIS_MAKE_SHORTSTRING("vpshrdq"),
    ZYDIS_MAKE_SHORTSTRING("vpshrdvd"),
    ZYDIS_MAKE_SHORTSTRING("vpshrdvq"),
    ZYDIS_MAKE_SHORTSTRING("vpshrdvw"),
    ZYDIS_MAKE_SHORTSTRING("vpshrdw"),
    ZYDIS_MAKE_SHORTSTRING("vpshufb"),
    ZYDIS_MAKE_SHORTSTRING("vpshufbitqmb"),
    ZYDIS_MAKE_SHORTSTRING("vpshufd"),
    ZYDIS_MAKE_SHORTSTRING("vpshufhw"),
    ZYDIS_MAKE_SHORTSTRING("vpshuflw"),
    ZYDIS_MAKE_SHORTSTRING("vpsignb"),
    ZYDIS_MAKE_SHORTSTRING("vpsignd"),
    ZYDIS_MAKE_SHORTSTRING("vpsignw"),
    ZYDIS_MAKE_SHORTSTRING("vpslld"),
    ZYDIS_MAKE_SHORTSTRING("vpslldq"),
    ZYDIS_MAKE_SHORTSTRING("vpsllq"),
    ZYDIS_MAKE_SHORTSTRING("vpsllvd"),
    ZYDIS_MAKE_SHORTSTRING("vpsllvq"),
    ZYDIS_MAKE_SHORTSTRING("vpsllvw"),
    ZYDIS_MAKE_SHORTSTRING("vpsllw"),
    ZYDIS_MAKE_SHORTSTRING("vpsrad"),
    ZYDIS_MAKE_SHORTSTRING("vpsraq"),
    ZYDIS_MAKE_SHORTSTRING("vpsravd"),
    ZYDIS_MAKE_SHORTSTRING("vpsravq"),
    ZYDIS_MAKE_SHORTSTRING("vpsravw"),
    ZYDIS_MAKE_SHORTSTRING("vpsraw"),
    ZYDIS_MAKE_SHORTSTRING("vpsrld"),
    ZYDIS_MAKE_SHORTSTRING("vpsrldq"),
    ZYDIS_MAKE_SHORTSTRING("vpsrlq"),
    ZYDIS_MAKE_SHORTSTRING("vpsrlvd"),
    ZYDIS_MAKE_SHORTSTRING("vpsrlvq"),
    ZYDIS_MAKE_SHORTSTRING("vpsrlvw"),
    ZYDIS_MAKE_SHORTSTRING("vpsrlw"),
    ZYDIS_MAKE_SHORTSTRING("vpsubb"),
    ZYDIS_MAKE_SHORTSTRING("vpsubd"),
    ZYDIS_MAKE_SHORTSTRING("vpsubq"),
    ZYDIS_MAKE_SHORTSTRING("vpsubrd"),
    ZYDIS_MAKE_SHORTSTRING("vpsubrsetbd"),
    ZYDIS_MAKE_SHORTSTRING("vpsubsb"),
    ZYDIS_MAKE_SHORTSTRING("vpsubsetbd"),
    ZYDIS_MAKE_SHORTSTRING("vpsubsw"),
    ZYDIS_MAKE_SHORTSTRING("vpsubusb"),
    ZYDIS_MAKE_SHORTSTRING("vpsubusw"),
    ZYDIS_MAKE_SHORTSTRING("vpsubw"),
    ZYDIS_MAKE_SHORTSTRING("vpternlogd"),
    ZYDIS_MAKE_SHORTSTRING("vpternlogq"),
    ZYDIS_MAKE_SHORTSTRING("vptest"),
    ZYDIS_MAKE_SHORTSTRING("vptestmb"),
    ZYDIS_MAKE_SHORTSTRING("vptestmd"),
    ZYDIS_MAKE_SHORTSTRING("vptestmq"),
    ZYDIS_MAKE_SHORTSTRING("vptestmw"),
    ZYDIS_MAKE_SHORTSTRING("vptestnmb"),
    ZYDIS_MAKE_SHORTSTRING("vptestnmd"),
    ZYDIS_MAKE_SHORTSTRING("vptestnmq"),
    ZYDIS_MAKE_SHORTSTRING("vptestnmw"),
    ZYDIS_MAKE_SHORTSTRING("vpunpckhbw"),
    ZYDIS_MAKE_SHORTSTRING("vpunpckhdq"),
    ZYDIS_MAKE_SHORTSTRING("vpunpckhqdq"),
    ZYDIS_MAKE_SHORTSTRING("vpunpckhwd"),
    ZYDIS_MAKE_SHORTSTRING("vpunpcklbw"),
    ZYDIS_MAKE_SHORTSTRING("vpunpckldq"),
    ZYDIS_MAKE_SHORTSTRING("vpunpcklqdq"),
    ZYDIS_MAKE_SHORTSTRING("vpunpcklwd"),
    ZYDIS_MAKE_SHORTSTRING("vpxor"),
    ZYDIS_MAKE_SHORTSTRING("vpxord"),
    ZYDIS_MAKE_SHORTSTRING("vpxorq"),
    ZYDIS_MAKE_SHORTSTRING("vrangepd"),
    ZYDIS_MAKE_SHORTSTRING("vrangeps"),
    ZYDIS_MAKE_SHORTSTRING("vrangesd"),
    ZYDIS_MAKE_SHORTSTRING("vrangess"),
    ZYDIS_MAKE_SHORTSTRING("vrcp14pd"),
    ZYDIS_MAKE_SHORTSTRING("vrcp14ps"),
    ZYDIS_MAKE_SHORTSTRING("vrcp14sd"),
    ZYDIS_MAKE_SHORTSTRING("vrcp14ss"),
    ZYDIS_MAKE_SHORTSTRING("vrcp23ps"),
    ZYDIS_MAKE_SHORTSTRING("vrcp28pd"),
    ZYDIS_MAKE_SHORTSTRING("vrcp28ps"),
    ZYDIS_MAKE_SHORTSTRING("vrcp28sd"),
    ZYDIS_MAKE_SHORTSTRING("vrcp28ss"),
    ZYDIS_MAKE_SHORTSTRING("vrcpph"),
    ZYDIS_MAKE_SHORTSTRING("vrcpps"),
    ZYDIS_MAKE_SHORTSTRING("vrcpsh"),
    ZYDIS_MAKE_SHORTSTRING("vrcpss"),
    ZYDIS_MAKE_SHORTSTRING("vreducepd"),
    ZYDIS_MAKE_SHORTSTRING("vreduceph"),
    ZYDIS_MAKE_SHORTSTRING("vreduceps"),
    ZYDIS_MAKE_SHORTSTRING("vreducesd"),
    ZYDIS_MAKE_SHORTSTRING("vreducesh"),
    ZYDIS_MAKE_SHORTSTRING("vreducess"),
    ZYDIS_MAKE_SHORTSTRING("vrndfxpntpd"),
    ZYDIS_MAKE_SHORTSTRING("vrndfxpntps"),
    ZYDIS_MAKE_SHORTSTRING("vrndscalepd"),
    ZYDIS_MAKE_SHORTSTRING("vrndscaleph"),
    ZYDIS_MAKE_SHORTSTRING("vrndscaleps"),
    ZYDIS_MAKE_SHORTSTRING("vrndscalesd"),
    ZYDIS_MAKE_SHORTSTRING("vrndscalesh"),
    ZYDIS_MAKE_SHORTSTRING("vrndscaless"),
    ZYDIS_MAKE_SHORTSTRING("vroundpd"),
    ZYDIS_MAKE_SHORTSTRING("vroundps"),
    ZYDIS_MAKE_SHORTSTRING("vroundsd"),
    ZYDIS_MAKE_SHORTSTRING("vroundss"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt14pd"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt14ps"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt14sd"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt14ss"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt23ps"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt28pd"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt28ps"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt28sd"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrt28ss"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrtph"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrtps"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrtsh"),
    ZYDIS_MAKE_SHORTSTRING("vrsqrtss"),
    ZYDIS_MAKE_SHORTSTRING("vscalefpd"),
    ZYDIS_MAKE_SHORTSTRING("vscalefph"),
    ZYDIS_MAKE_SHORTSTRING("vscalefps"),
    ZYDIS_MAKE_SHORTSTRING("vscalefsd"),
    ZYDIS_MAKE_SHORTSTRING("vscalefsh"),
    ZYDIS_MAKE_SHORTSTRING("vscalefss"),
    ZYDIS_MAKE_SHORTSTRING("vscaleps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterdpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterdps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf0dpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf0dps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf0hintdpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf0hintdps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf0qpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf0qps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf1dpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf1dps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf1qpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterpf1qps"),
    ZYDIS_MAKE_SHORTSTRING("vscatterqpd"),
    ZYDIS_MAKE_SHORTSTRING("vscatterqps"),
    ZYDIS_MAKE_SHORTSTRING("vshuff32x4"),
    ZYDIS_MAKE_SHORTSTRING("vshuff64x2"),
    ZYDIS_MAKE_SHORTSTRING("vshufi32x4"),
    ZYDIS_MAKE_SHORTSTRING("vshufi64x2"),
    ZYDIS_MAKE_SHORTSTRING("vshufpd"),
    ZYDIS_MAKE_SHORTSTRING("vshufps"),
    ZYDIS_MAKE_SHORTSTRING("vsqrtpd"),
    ZYDIS_MAKE_SHORTSTRING("vsqrtph"),
    ZYDIS_MAKE_SHORTSTRING("vsqrtps"),
    ZYDIS_MAKE_SHORTSTRING("vsqrtsd"),
    ZYDIS_MAKE_SHORTSTRING("vsqrtsh"),
    ZYDIS_MAKE_SHORTSTRING("vsqrtss"),
    ZYDIS_MAKE_SHORTSTRING("vstmxcsr"),
    ZYDIS_MAKE_SHORTSTRING("vsubpd"),
    ZYDIS_MAKE_SHORTSTRING("vsubph"),
    ZYDIS_MAKE_SHORTSTRING("vsubps"),
    ZYDIS_MAKE_SHORTSTRING("vsubrpd"),
    ZYDIS_MAKE_SHORTSTRING("vsubrps"),
    ZYDIS_MAKE_SHORTSTRING("vsubsd"),
    ZYDIS_MAKE_SHORTSTRING("vsubsh"),
    ZYDIS_MAKE_SHORTSTRING("vsubss"),
    ZYDIS_MAKE_SHORTSTRING("vtestpd"),
    ZYDIS_MAKE_SHORTSTRING("vtestps"),
    ZYDIS_MAKE_SHORTSTRING("vucomisd"),
    ZYDIS_MAKE_SHORTSTRING("vucomish"),
    ZYDIS_MAKE_SHORTSTRING("vucomiss"),
    ZYDIS_MAKE_SHORTSTRING("vunpckhpd"),
    ZYDIS_MAKE_SHORTSTRING("vunpckhps"),
    ZYDIS_MAKE_SHORTSTRING("vunpcklpd"),
    ZYDIS_MAKE_SHORTSTRING("vunpcklps"),
    ZYDIS_MAKE_SHORTSTRING("vxorpd"),
    ZYDIS_MAKE_SHORTSTRING("vxorps"),
    ZYDIS_MAKE_SHORTSTRING("vzeroall"),
    ZYDIS_MAKE_SHORTSTRING("vzeroupper"),
    ZYDIS_MAKE_SHORTSTRING("wbinvd"),
    ZYDIS_MAKE_SHORTSTRING("wrfsbase"),
    ZYDIS_MAKE_SHORTSTRING("wrgsbase"),
    ZYDIS_MAKE_SHORTSTRING("wrmsr"),
    ZYDIS_MAKE_SHORTSTRING("wrpkru"),
    ZYDIS_MAKE_SHORTSTRING("wrssd"),
    ZYDIS_MAKE_SHORTSTRING("wrssq"),
    ZYDIS_MAKE_SHORTSTRING("wrussd"),
    ZYDIS_MAKE_SHORTSTRING("wrussq"),
    ZYDIS_MAKE_SHORTSTRING("xabort"),
    ZYDIS_MAKE_SHORTSTRING("xadd"),
    ZYDIS_MAKE_SHORTSTRING("xbegin"),
    ZYDIS_MAKE_SHORTSTRING("xchg"),
    ZYDIS_MAKE_SHORTSTRING("xcrypt_cbc"),
    ZYDIS_MAKE_SHORTSTRING("xcrypt_cfb"),
    ZYDIS_MAKE_SHORTSTRING("xcrypt_ctr"),
    ZYDIS_MAKE_SHORTSTRING("xcrypt_ecb"),
    ZYDIS_MAKE_SHORTSTRING("xcrypt_ofb"),
    ZYDIS_MAKE_SHORTSTRING("xend"),
    ZYDIS_MAKE_SHORTSTRING("xgetbv"),
    ZYDIS_MAKE_SHORTSTRING("xlat"),
    ZYDIS_MAKE_SHORTSTRING("xor"),
    ZYDIS_MAKE_SHORTSTRING("xorpd"),
    ZYDIS_MAKE_SHORTSTRING("xorps"),
    ZYDIS_MAKE_SHORTSTRING("xresldtrk"),
    ZYDIS_MAKE_SHORTSTRING("xrstor"),
    ZYDIS_MAKE_SHORTSTRING("xrstor64"),
    ZYDIS_MAKE_SHORTSTRING("xrstors"),
    ZYDIS_MAKE_SHORTSTRING("xrstors64"),
    ZYDIS_MAKE_SHORTSTRING("xsave"),
    ZYDIS_MAKE_SHORTSTRING("xsave64"),
    ZYDIS_MAKE_SHORTSTRING("xsavec"),
    ZYDIS_MAKE_SHORTSTRING("xsavec64"),
    ZYDIS_MAKE_SHORTSTRING("xsaveopt"),
    ZYDIS_MAKE_SHORTSTRING("xsaveopt64"),
    ZYDIS_MAKE_SHORTSTRING("xsaves"),
    ZYDIS_MAKE_SHORTSTRING("xsaves64"),
    ZYDIS_MAKE_SHORTSTRING("xsetbv"),
    ZYDIS_MAKE_SHORTSTRING("xsha1"),
    ZYDIS_MAKE_SHORTSTRING("xsha256"),
    ZYDIS_MAKE_SHORTSTRING("xstore"),
    ZYDIS_MAKE_SHORTSTRING("xsusldtrk"),
    ZYDIS_MAKE_SHORTSTRING("xtest")
};
